from __future__ import print_function
import random
import json
from subprocess import check_output, CalledProcessError, STDOUT
import requests
import time
import datetime
import os
import pkg_resources

from bitcoin import serialize
from concurrent import futures

__version__ = pkg_resources.get_distribution('moneywagon').version
useragent = "Moneywagon %s" % __version__

class ServiceDisagreement(Exception):
    pass

class NoService(Exception):
    pass

class SkipThisService(NoService):
    pass

class ServiceError(NoService):
    pass

class CurrencyNotSupported(Exception):
    pass

class NoServicesDefined(Exception):
    pass

class NoData(Exception):
    pass

class RevertToPrivateMode(NotImplementedError):
    pass

class ClassProperty(property):
    """
    From http://stackoverflow.com/a/1383402/118495
    """
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()

class Service(object):
    """
    Represents a blockchain service running an Http interface.
    Some `Services` subclass will only support a subset of all pissible blockchain functions.
    All Services should subclass this class, and implement their own `get_*` method.
    """
    domain = None # optional, useful if the service will have subclasses
    api_homepage = '' # link to page defining the API.
    protocol = 'https'
    supported_cryptos = None # must be a list of lower case currency codes.
    explorer_address_url = None # url to block explerer page. Use {address} and {crypto} as placeholders.
    explorer_tx_url = None # {txid}
    explorer_blocknum_url = None # {blocknum}
    explorer_blockhash_url = None # {blockhash}
    ssl_verify = True
    socketio_url = None
    exchange_fee_rate = None
    api_key = False
    symbol_mapping = None

    @ClassProperty
    @classmethod
    def name(cls):
        return cls.__name__

    def __init__(self, verbose=False, responses=None, timeout=None, random_wait_seconds=0, api_key=None, api_secret=None):
        self.responses = responses or {} # for caching
        self.verbose = verbose
        self.last_url = None
        self.last_raw_response = None
        self.timeout = timeout
        self.random_wait_seconds = random_wait_seconds
        self.api_key = api_key
        self.api_secret = api_secret
        self.total_external_fetch_duration = datetime.timedelta(0)

        try:
            with open(os.path.expanduser('~/.exchange_keys')) as f:
                config = json.loads(f.read())

            for key, value in config[self.name].items():
                if not hasattr(self, key) or not getattr(self, key):
                    # only load if no other values have been passed in.
                    setattr(self, key, str(value))
        except KeyError:
            pass # no key found
        except Exception as exc:
            if verbose:
                print("config file broke", str(exc))


    def __repr__(self):
        return "<Service: %s (%s in cache)>" % (self.__class__.__name__, len(self.responses))

    def get_url(self, url, *args, **kwargs):
        return self._external_request('get', url, *args, **kwargs)

    def post_url(self, url, *args, **kwargs):
        return self._external_request('post', url, *args, **kwargs)

    def delete_url(self, url, *args, **kwargs):
        return self._external_request('delete', url, *args, **kwargs)

    def check_error(self, response):
        """
        If the service is returning an error, this function should raise an exception.
        such as SkipThisService
        """
        if response.status_code == 500:
            raise ServiceError("500 - " + response.content)

        if response.status_code == 503:
            if "DDoS protection by Cloudflare" in response.content:
                raise ServiceError("Foiled by Cloudfare's DDoS protection")
            raise ServiceError("503 - Temporarily out of service.")

        if response.status_code == 429:
            raise ServiceError("429 - Too many requests")

        if response.status_code == 404:
            raise ServiceError("404 - Not Found")

        if response.status_code == 400:
            raise ServiceError("400 - Bad Request")

    def convert_currency(self, base_fiat, base_amount, target_fiat):
        """
        Convert one fiat amount to another fiat. Uses the fixer.io service.
        """
        url = "http://api.fixer.io/latest?base=%s" % base_fiat
        data = self.get_url(url).json()
        try:
            return data['rates'][target_fiat.upper()] * base_amount
        except KeyError:
            raise Exception("Can not convert %s to %s" % (base_fiat, target_fiat))

    def make_market(self, crypto, fiat):
        """
        Make market identifier the service can understand with the passed in
        moneywagon format crypto and fiat. Usually calls `self.fix_symbol`.
        """
        raise NotImplementedError()

    def fix_symbol(self, symbol, reverse=False):
        """
        In comes a moneywagon format symbol, and returned in the symbol converted
        to one the service can understand.
        """
        if not self.symbol_mapping:
            return symbol

        for old, new in self.symbol_mapping:
            if reverse:
                if symbol == new:
                    return old
            else:
                if symbol == old:
                    return new

        return symbol

    def parse_market(self, market, split_char='_'):
        """
        In comes the market identifier directly from the service. Returned is
        the crypto and fiat identifier in moneywagon format.
        """
        crypto, fiat = market.lower().split(split_char)
        return (
            self.fix_symbol(crypto, reverse=True),
            self.fix_symbol(fiat, reverse=True)
        )

    def make_market(self, crypto, fiat):
        """
        Convert a crypto and fiat to a "market" string. All exchanges use their
        own format for specifying markets. Subclasses can define their own
        implementation.
        """
        return ("%s_%s" % (self.fix_symbol(crypto), self.fix_symbol(fiat))).lower()

    def make_rpc_call(self, args, internal=False, skip_json=False):
        cmd = [self.cli_path] + [str(a) for a in args]
        if not self.cli_path:
            raise SkipThisService("No CLI path defined")
        try:
            raw = check_output(cmd, stderr=STDOUT)
        except CalledProcessError as exc:
            raise SkipThisService("Full node CLI call failed: %s" % exc.output)
        except OSError as exc:
            raise SkipThisService(str(exc))

        if not internal:
            self.last_raw_response = raw
            self.last_url = " ".join(cmd)

        ret = raw
        if not skip_json:
            ret = json.loads(raw)

        return ret

    def _external_request(self, method, url, *args, **kwargs):
        """
        Wrapper for requests.get with useragent automatically set.
        And also all requests are reponses are cached.
        """
        self.last_url = url
        if url in self.responses.keys() and method == 'get':
            return self.responses[url] # return from cache if its there

        headers = kwargs.pop('headers', None)
        custom = {'User-Agent': useragent}
        if headers:
            headers.update(custom)
            kwargs['headers'] = headers
        else:
            kwargs['headers'] = custom

        if self.timeout:
            # add timeout parameter to requests.get if one was passed in on construction...
            kwargs['timeout'] = self.timeout

        start = datetime.datetime.now()
        response = getattr(requests, method)(url, verify=self.ssl_verify, *args, **kwargs)
        self.total_external_fetch_duration += datetime.datetime.now() - start

        if self.verbose:
            print("Got Response: %s (took %s)" % (url, (datetime.datetime.now() - start)))

        self.last_raw_response = response

        self.check_error(response)

        if method == 'get':
            self.responses[url] = response # cache for later

        return response

    def get_current_price(self, crypto, fiat):
        """
        Makes call to external service, and returns the price for given
        fiat/crypto pair. Returned is a float.
        """
        raise NotImplementedError(
            self.name + " does not support getting the current fiat exchange rate. "
            "Or rather it has no defined 'get_current_price' method."
        )

    def get_historical_price(self, crypto, fiat, at_time):
        """
        """
        raise NotImplementedError(
            self.name + " does not support getting historical price. "
            "Or rather it has no defined 'get_historical_price' method."
        )

    def get_transactions(self, crypto, address, confirmations=1):
        """
        Must be returned with the most recent transaction at the top.
        Returned is a list of dicts that have the following keys:

        `amount`: Number of units of currency moved. Always in base units (not satoshis).
        `date`: a datetime object of when the transaction was made.
        `txid`: The transaction ID, looks like a hash.
        `confirmations`: integer of the number of confirmations this transaction has on top of it.

        """
        raise NotImplementedError(
            self.name + " does not support getting historical transactions. "
            "Or rather it has no defined 'get_transactions' method."
        )

    def get_transactions_multi(self, crypto, addresses, confirmations=1):
        raise NotImplementedError(
            self.name + " does not support getting historical transactions by multiple addresses. "
            "Or rather it has no defined 'get_transactions_multi' method."
        )

    def get_single_transaction(self, crypto, txid):
        """
        Get detailed information about a single transaction.

        time - datetime, when this transaction was confirmed.
        block_hash - string, the id of the block this tx is confirmed in.
        block_number - integer, which block number this tx is confirmed in.
        hex - the entire tx encoded in hex format
        size - size of TX in bytes.
        inputs - list of {address:, amount:, txid:}, amount is in satoshis
        outputs - list of {address:, amount:, scriptPubKey:}, amount is in satoshis, scriptPubKey is hex
        txid
        total_out - (in satoshis)
        total_ins - (in satoshis)
        confirmations - number of confirmations this TX curently has
        fee - total amount of fees this TX leaves for miners (in satoshis)

        """
        raise NotImplementedError(
            self.name + " does not support getting single transactions. "
            "Or rather it has no defined 'get_single_transaction' method."
        )

    def get_single_transaction_multi(self, crypto, txids):
        raise NotImplementedError(
            self.name + " does not support getting single transactions by multiple txids. "
            "Or rather it has no defined 'get_single_transaction_multi' method."
        )


    def get_unspent_outputs(self, crypto, address):
        """
        Default implmentation of this function that uses get_transaction
        Subclasses should overwrite this with a direct call to get utxo (if applicable)

        Returned will be a list of dictionaries, the keys will be:

        required:

        `output` - the big endian tx hash, followed by a colon, then the tx index. (for pybitcointools support)
        `address` - the address passed in (for pybitcointools support)
        `amount` - int, satoshi amount of the input
        `scriptPubKey` - string
        `txid` - string
        `vout` - integer

        optional:

        `confirmations` - how many confirmations this tx has so far.
        `scriptPubKey_asm` - string

        """
        raise NotImplementedError(
            self.name + " does not support getting unspent outputs. "
            "Or rather it has no defined 'get_unspent_outputs' method."
        )

    def get_unspent_outputs_multi(self, crypto, addresses):
        raise NotImplementedError(
            self.name + " does not support getting unspent outputs by multiple addresses. "
            "Or rather it has no defined 'get_unspent_outputs_multi' method."
        )

    def get_balance(self, crypto, address, confirmations=1):
        """
        Get the amount of coin in the address passed in.
        Always returns a single float.
        """
        raise NotImplementedError(
            self.name + " does not support getting address balances. "
            "Or rather it has no defined 'get_balance' method."
        )

    def get_balance_multi(self, crypto, addresses, confirmations=1):
        """
        Same as above, except addresses are passed in as a list instead of
        just a single string.
        """
        raise NotImplementedError(
            self.name + " does not support getting multiple address balances. "
            "Or rather it has no defined 'get_balance_multi' method."
        )

    def push_tx(self, crypto, tx_hex):
        """
        Push transaction to the miner network. Returns txid if done
        successfully.
        """
        raise NotImplementedError(
            self.name + " does not support pushing transactions to the network. "
            "Or rather it has no defined 'push_tx' method."
        )

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        """
        Get block based on either block height, block number or get the latest
        block. Only one of the previous arguments must be passed on.

        Returned is a dictionary object with the following keys:

        * required fields:

        block_number - int
        size - size of block
        time - datetime object of when the block was made
        hash - str (must be all lowercase)
        tx_count - int, the number of transactions included in thi block.

        * optional fields:

        confirmations - int
        sent_value - total value moved from all included transactions
        total_fees - total amount of tx fees from all included transactions
        mining_difficulty - what the difficulty was when this block was made.
        merkle_root - str (lower case)
        previous_hash - str (lower case)
        next_hash - str (lower case) (or `None` of its the latest block)
        miner - name of miner that first relayed this block
        version - block version
        """
        raise NotImplementedError(
            self.name + " does not support getting getting block data. "
            "Or rather it has no defined 'get_block' method."
        )

    def get_optimal_fee(self, crypto, tx_bytes):
        raise NotImplementedError(
            self.name + " does not support getting optimal fee. "
            "Or rather it has no defined 'get_optimal_fee' method."
        )

    def get_pairs(self):
        """
        Only for exchanges. Returns list of all trading pairs supported by this
        service. All items are lower case, and seperated with a "-".
        """
        raise NotImplementedError(
            self.name + " does not support getting all exchange pairs. "
            "Or rather it has no defined 'get_pairs' method."
        )

    def get_orderbook(self, crypto, fiat):
        """
        Only for exchznges. Returns the current orderbook as a dict with two keys:
        bids and asks. Each is a list of floats. The first float is price, the
        second is order size.
        """
        raise NotImplementedError(
            self.name + " does not support getting orderbook. "
            "Or rather it has no defined 'get_orderbook' method."
        )

    def make_order(self, crypto, fiat, amount, price, type="limit"):
        """
        This method buys or sells `crypto` on an exchange using `fiat` balance.
        Type can either be "fill-or-kill", "post-only", "market", or "limit".
        To get what modes are supported, consult make_order.supported_types
        if one is defined.
        """
        raise NotImplementedError(
            self.name + " does not support making orders. "
            "Or rather it has no defined 'make_order' method."
        )

    def cancel_order(self, order_id):
        raise NotImplementedError(
            self.name + " does not support canceling orders. "
            "Or rather it has no defined 'cancel_order' method."
        )

    def list_orders(self, status="open"):
        raise NotImplementedError(
            self.name + " does not support listing orders. "
            "Or rather it has no defined 'list_orders' method."
        )

    def initiate_withdraw(self, crypto, amount, address):
         raise NotImplementedError(
             self.name + " does not support initiating withdraws. "
             "Or rather it has no defined 'initiate_withdraw' method."
         )

    def get_deposit_address(self, crypto):
        raise NotImplementedError(
            self.name + " does not support getting deposit addresses. "
            "Or rather it has no defined 'get_deposit_address' method."
        )

    def get_exchange_balance(self, crypto, type="available"):
        """
        Gives you the balance you have on that exchange. Not to be confused with
        `get_balance`, which is for address balance. Returned is a float.
        """
        raise NotImplementedError(
            self.name + " does not support getting exchange balance. "
            "Or rather it has no defined 'get_exchange_balance' method."
        )

    def get_total_exchange_balances(self):
        """
        Returns balances for all currencies held on this service. Returned is a
        dictionary with currency code as key and float as the value.
        """
        raise NotImplementedError(
            self.name + " does not support getting total exchange balances. "
            "Or rather it has no defined 'get_total_exchange_balances' method."
        )

    def generate_new_deposit_address(self, crypto):
        raise NotImplementedError(
            self.name + " does not support generating new deposit address. "
            "Or rather it has no defined 'generate_new_deposit_address' method."
        )

class AutoFallbackFetcher(object):
    """
    Calls a succession of services until one returns a value.
    """

    def __init__(self, services=None, verbose=False, responses=None, timeout=None, random_wait_seconds=0):
        """
        Each service class is instantiated here so the service instances stay
        in scope for the entire life of this object. This way the service
        objects can cache responses.
        """
        if not services:
            from moneywagon import ALL_SERVICES
            services = ALL_SERVICES

        self.services = []
        for ServiceClass in services:
            self.services.append(
                ServiceClass(verbose=verbose, responses=responses, timeout=timeout)
            )

        self.verbose = verbose
        self._successful_service = None # gets filled in after success
        self._failed_services = []
        self.random_wait_seconds = random_wait_seconds

    def _try_services(self, method_name, *args, **kwargs):
        """
        Try each service until one returns a response. This function only
        catches the bare minimum of exceptions from the service class. We want
        exceptions to be raised so the service classes can be debugged and
        fixed quickly.
        """
        crypto = ((args and args[0]) or kwargs['crypto']).lower()
        address = kwargs.get('address', '').lower()
        fiat = kwargs.get('fiat', '').lower()

        if not self.services:
            raise CurrencyNotSupported("No services defined for %s for %s" % (method_name, crypto))

        if self.random_wait_seconds > 0:
            # for privacy... To avoid correlating addresses to same origin
            # only gets called before the first service call. Does not pause
            # before each and every call.
            pause_time = random.random() * self.random_wait_seconds
            if self.verbose:
                print("Pausing for: %.2f seconds" % pause_time)
            time.sleep(pause_time)

        for service in self.services:
            if service.supported_cryptos and (crypto not in service.supported_cryptos):
                if self.verbose:
                    print("SKIP:", "%s not supported for %s" % (crypto, service.__class__.__name__))
                continue
            try:
                if self.verbose: print("* Trying:", service, crypto, "%s%s" % (address, fiat))
                ret =  getattr(service, method_name)(*args, **kwargs)
                self._successful_service = service
                return ret
            except (KeyError, IndexError, TypeError, ValueError,
                    requests.exceptions.Timeout, requests.exceptions.ConnectionError) as exc:
                # API has probably changed, therefore service class broken
                if self.verbose: print("FAIL:", service, exc.__class__.__name__, exc)
                self._failed_services.append({
                    'service': service,
                    'error': "%s %s" % (exc.__class__.__name__, exc)
                })
            except NoService as exc:
                # service classes can raise this exception if for whatever reason
                # that service can't return a response, but maybe another one can.
                if self.verbose: print("SKIP:", exc.__class__.__name__, exc)
                self._failed_services.append({'service': service, 'error': "Skipped: %s" % str(exc)})
            except NotImplementedError as exc:
                if self.verbose: print("SKIP:", exc.__class__.__name__, exc)
                self._failed_services.append({'service': service, 'error': "Not Implemented"})


        if not self._failed_services:
            raise NotImplementedError(
                "No Services defined for %s and %s" % (crypto, method_name)
            )

        if set(x['error'] for x in self._failed_services) == set(['Not Implemented']) and method_name.endswith("multi"):
            # some currencies may not have any multi functions defined, so retry
            # with private mode (which tries multiple services).
            raise RevertToPrivateMode("All services do not implement %s service" % method_name)

        failed_msg = ', '.join(
            ["{service.name} -> {error}".format(**x) for x in self._failed_services]
        )
        raise NoService(self.no_service_msg(*args, **kwargs) + "! Tried: " + failed_msg)

    def no_service_msg(self, *args, **kwargs):
        """
        This function is called when all Services have been tried and no value
        can be returned. It much take the same args and kwargs as in the method
        spefified in `method_name`. Returned is a string for the error message.
        It should say something informative.
        """
        return "All either skipped or failed."


def enforce_service_mode(services, FetcherClass, kwargs, modes):
    """
    Fetches the value according to the mode of execution desired.
    `FetcherClass` must be a class that is subclassed from AutoFallbackFetcher.
    `services` must be a list of Service classes.
    `kwargs` is a list of arguments used to make the service call, usually
      something like {crypto: 'btc', address: '1HwY...'} or
      {crypto: 'ltc', fiat: 'rur'}, (depends on the what FetcherClass.action takes)

    Modes can be:

         average = positive int. 1 by default. Takes average of n fetches.
         verbose = [True|False] False by default. Extra output.
         random = [True|False] False by default. Randomizes service order.
         paranoid = positive int. 1 by default. Redundant Fetching.
         fast = positive int. 0 by default. Return as soon as recieved first n results.

    """
    fast_level = modes.get('fast', 0)
    average_level = modes.get('average', 0)
    paranoid_level = modes.get('paranoid', 0)
    private_level = modes.get('private', 0)
    verbose = modes.get('verbose', False)
    timeout = modes.get('timeout', None)

    if len(services) == 0:
        raise NoService("No services defined")

    if modes.get('random', False):
        random.shuffle(services)

    if private_level  > 0:
        results = _do_private_mode(
            FetcherClass, services, kwargs, random_wait_seconds=private_level,
            verbose=verbose, timeout=timeout
        )
        return results

    elif average_level <= 1 and paranoid_level <= 1 and fast_level == 0:
        # only need to make 1 external call, no need for threading...
        fetcher = FetcherClass(services=services, verbose=verbose, timeout=timeout)
        consensus_results = fetcher.action(**kwargs)
        used_services = [fetcher._successful_service]

    elif average_level > 1:
        # instead of checking that all results are the same, we just return the average of all results.
        # mostly useful for non-blockchain operations like price and optimal fee.
        results = _get_results(
            FetcherClass, services, kwargs, num_results=average_level, verbose=verbose, timeout=timeout
        )

        to_compare, used_services = _prepare_consensus(FetcherClass, results)
        consensus_results = sum(to_compare) / len(to_compare)

    elif paranoid_level > 1:
        results = _get_results(
            FetcherClass, services, kwargs, num_results=paranoid_level, verbose=verbose, timeout=timeout
        )
        to_compare, used_services = _prepare_consensus(FetcherClass, results)

        if len(set(to_compare)) > 1:
            full_results = zip(used_services, to_compare)
            show_error = ", ".join("%s: %s" % (s.name, v) for s, v in full_results)
            sd = ServiceDisagreement("No service consensus: %s" % show_error)
            sd.network_results = {
                service: result for service, result in full_results
            }
            raise sd

        # if all values match, return any one (in this case the first one).
        # also return the list of all services that confirm this result.
        consensus_results = results[0][1]

    elif fast_level >= 1:
        raise Exception("Fast mode not yet working")
        results = _get_results(
            FetcherClass, services, kwargs, fast=fast_level, verbose=verbose
        )

    else:
        raise Exception("No mode specified.")

    if modes.get('report_services'):
        return used_services, consensus_results
    else:
        return consensus_results


def _prepare_consensus(FetcherClass, results):
    """
    Given a list of results, return a list that is simplified to make consensus
    determination possible. Returns two item tuple, first arg is simplified list,
    the second argument is a list of all services used in making these results.
    """
    # _get_results returns lists of 2 item list, first element is service, second is the returned value.
    # when determining consensus amoung services, only take into account values returned.
    if hasattr(FetcherClass, "strip_for_consensus"):
        to_compare = [
            FetcherClass.strip_for_consensus(value) for (fetcher, value) in results
        ]
    else:
        to_compare = [value for fetcher, value in results]

    return to_compare, [fetcher._successful_service for fetcher, values in results]

def _get_results(FetcherClass, services, kwargs, num_results=None, fast=0, verbose=False, timeout=None):
    """
    Does the fetching in multiple threads of needed. Used by paranoid and fast mode.
    """
    results = []

    if not num_results or fast:
        num_results = len(services)

    with futures.ThreadPoolExecutor(max_workers=len(services)) as executor:
        fetches = {}
        for service in services[:num_results]:
            tail = [x for x in services if x is not service]
            random.shuffle(tail)
            srv = FetcherClass(services=[service] + tail, verbose=verbose, timeout=timeout)
            fetches[executor.submit(srv.action, **kwargs)] = srv

        if fast == 1:
            raise NotImplementedError
            # ths code is a work in progress. futures.FIRST_COMPLETED works differently than I thought...
            to_iterate, still_going = futures.wait(fetches, return_when=futures.FIRST_COMPLETED)
            for x in still_going:
                try:
                    x.result(timeout=1.001)
                except futures._base.TimeoutError:
                    pass

        elif fast > 1:
            raise Exception("fast level greater than 1 not yet implemented")
        else:
            to_iterate = futures.as_completed(fetches)

        for future in to_iterate:
            service = fetches[future]
            results.append([service, future.result()])

    return results

def _do_private_mode(FetcherClass, services, kwargs, random_wait_seconds, timeout, verbose):
    """
    Private mode is only applicable to address_balance, unspent_outputs, and
    historical_transactions. There will always be a list for the `addresses`
    argument. Each address goes to a random service. Also a random delay is
    performed before the external fetch for improved privacy.
    """
    addresses = kwargs.pop('addresses')
    results = {}

    with futures.ThreadPoolExecutor(max_workers=len(addresses)) as executor:
        fetches = {}
        for address in addresses:
            k = kwargs
            k['address'] = address
            random.shuffle(services)
            srv = FetcherClass(
                services=services, verbose=verbose, timeout=timeout or 5.0,
                random_wait_seconds=random_wait_seconds
            )
            # address is returned because balance needs to be returned
            # attached to the address. Other methods (get_transaction, unspent_outputs, etc)
            # do not need to be indexed by address. (upstream they are stripped out)
            fetches[executor.submit(srv.action, **k)] = (srv, address)

        to_iterate = futures.as_completed(fetches)

        for future in to_iterate:
            service, address = fetches[future]
            results[address] = future.result()

    return results


def currency_to_protocol(amount):
    """
    Convert a string of 'currency units' to 'protocol units'. For instance
    converts 19.1 bitcoin to 1910000000 satoshis.

    Input is a float, output is an integer that is 1e8 times larger.

    It is hard to do this conversion because multiplying
    floats causes rounding nubers which will mess up the transactions creation
    process.

    examples:

    19.1 -> 1910000000
    0.001 -> 100000

    """
    if type(amount) in [float, int]:
        amount = "%.8f" % amount

    return int(amount.replace(".", '')) # avoiding float math

def get_optimal_services(crypto, type_of_service):
    from .crypto_data import crypto_data
    try:
        # get best services from curated list
        services = crypto_data[crypto.lower()]['services']
    except (KeyError, TypeError):
        raise CurrencyNotSupported("Unknown cryptocurrency symbol: %s" % crypto)

    try:
        s = services[type_of_service]
        if not s:
            raise KeyError()
        return s
    except KeyError:
        raise NoServicesDefined("No %s services defined for %s" % (type_of_service, crypto))


def get_magic_bytes(crypto):
    from .crypto_data import crypto_data
    try:
        pub_byte = crypto_data[crypto]['address_version_byte']
        priv_byte = crypto_data[crypto]['private_key_prefix']
        if priv_byte >= 128:
            priv_byte -= 128 #pybitcointools bug
        return pub_byte, priv_byte

    except KeyError:
        raise ValueError("Cryptocurrency symbol not found: %s" % crypto)

def make_standard_halfing_eras(start, interval, start_reward, halfing_func=None, total_eras=10):
    if not halfing_func:
        halfing_func = lambda x, y: y / float(2 ** (x))
    return [
    {
        'start': start + (era * interval) + 1,
        'end': start + ((era + 1) * interval),
        'reward': halfing_func(era, start_reward)
    } for era in range(0, total_eras)
]

def decompile_scriptPubKey(asm):
    """
    >>> decompile_scriptPubKey('OP_DUP OP_HASH160 cef3550ff9e637ddd120717d43fc21f8a563caf8 OP_EQUALVERIFY OP_CHECKSIG')
    '76a914cef3550ff9e637ddd120717d43fc21f8a563caf888ac'
    """
    asm = asm.split(" ")
    hex = ""
    if asm[0] == 'OP_DUP':
        hex += "76"
    if asm[1] == 'OP_HASH160':
        hex += 'a9'
    if len(asm[2]) == 40:
        hex += asm[2]
    if asm[3] == 'OP_EQUALVERIFY':
        hex += '88'
    if asm[4] == 'OP_CHECKSIG':
        hex += 'ac'
    return hex

def to_rawtx(tx):
    """
    Take a tx object in the moneywagon format and convert it to the format
    that pybitcointools's `serialize` funcion takes, then return in raw hex
    format.
    """
    if tx.get('hex'):
        return tx['hex']

    new_tx = {}

    locktime = tx.get('locktime', 0)
    new_tx['locktime'] = locktime
    new_tx['version'] = tx.get('version', 1)
    new_tx['ins'] = [
        {
            'outpoint': {'hash': str(x['txid']), 'index': x['n']},
            'script': str(x['scriptSig'].replace(' ', '')),
            'sequence': x.get('sequence', 0xFFFFFFFF if locktime == 0 else None)
        } for x in tx['inputs']
    ]

    new_tx['outs'] = [
        {
            'script': str(x['scriptPubKey']),
            'value': x['amount']
        } for x in tx['outputs']
    ]

    return serialize(new_tx)



import requests
import re

def get_content_from_github(github_path, file):
    url = "https://raw.githubusercontent.com/%s/master/src/%s" % (github_path, file)
    response = requests.get(url)
    if response.status_code != 200:
        return None
    return response.content

def extract_crypto_data(github_path):
    """
    github_path can must be path on github, such as
    "bitcoin/bitcoin" or "litecoin-project/litecoin"
    """
    data = {'github_link': 'https://github.com/%s' % github_path}

    content = get_content_from_github(github_path, "chainparams.cpp")
    if content:
        data.update(_get_from_chainparams(content))
    else:
        content = get_content_from_github(github_path, "base58.h")
        if content:
            data.update(_get_from_base58h(content))

    content = get_content_from_github(github_path, "main.cpp")
    if content:
        data.update(_get_from_main(content))

    return data

def _get_from_main(content):
    data = {}
    match = test_regexes(content,
        "if\s+\(txPrev\.nTime\s+>\s+nTime\)", # peercoin style PoS
        "if\s+\(\(unsigned int\)coins\.nTime\s+>\s+tx\.nTime\)" # rdd style PoSV
    )
    if match:
        data['transaction_form'] = 'ppc-timestamp'

    return data

def _get_from_base58h(content):
    data = {}
    m = re.search("PUBKEY_ADDRESS = (\d+),", content)
    if m:
        data['address_version_byte'] = int(m.groups()[0])

    if "PRIVKEY_ADDRESS = CBitcoinAddress::PUBKEY_ADDRESS + 128," in content:
        data['private_key_prefix'] = data['address_version_byte'] + 128

    if "SetData(128 + (fTestNet ? CBitcoinAddress::PUBKEY_ADDRESS_TEST : CBitcoinAddress::PUBKEY_ADDRESS), &vchSecret[0], vchSecret.size());" in content:
        data['private_key_prefix'] = data['address_version_byte'] + 128

    m = re.search("SCRIPT_ADDRESS = (\d+), ", content)
    if m:
        data['script_hash_byte'] = int(m.groups()[0])

    return data


def _get_from_chainparams(content):
    # newer style "chain params" definitions.
    data = {}
    match = test_regexes(content,
        "base58Prefixes\[PUBKEY_ADDRESS\]\s+=\s+std::vector<unsigned char>\(1,(\d+)\);",
        "base58Prefixes\[PUBKEY_ADDRESS\]\s+=\s+list_of\((\d+)\).convert_to_container<std::vector<unsigned char> >\(\);",
        "base58Prefixes\[PUBKEY_ADDRESS\]\s+=\s+list_of\((\d+)\);"
    )
    if match:
        data['address_version_byte'] = int(match)

    match = test_regexes(content,
        "base58Prefixes\[SCRIPT_ADDRESS\]\s+=\s+std::vector<unsigned char>\(1,(\d+)\);",
        "base58Prefixes\[SCRIPT_ADDRESS\]\s+=\s+list_of\((\d+)\).convert_to_container<std::vector<unsigned char> >\(\);",
        "base58Prefixes\[SCRIPT_ADDRESS\]\s+=\s+list_of\((\d+)\);"
    )
    if match:
        data['script_hash_byte'] = int(match)

    match = test_regexes(content,
        "base58Prefixes\[SECRET_KEY\]\s+=\s+std::vector<unsigned char>\(1,(\d+)\);",
        "base58Prefixes\[SECRET_KEY\]\s+=\s+list_of\((\d+)\).convert_to_container<std::vector<unsigned char> >\(\);",
        "base58Prefixes\[SECRET_KEY\]\s+=\s+list_of\((\d+)\);"
    )
    if match:
        data['private_key_prefix'] = int(match)

    data['message_magic'] = b""
    for index in range(4):
        m = re.search("pchMessageStart\[%s\] = 0x([\d\w]+);" % index, content)
        data['message_magic'] += bytes(m.groups()[0])

    data['seed_nodes'] = []
    seed_index = 0
    # newest seed definition style
    m = re.findall('vSeeds.emplace_back\("([\w\d.]+)", (true|false)\);', content)

    if not m:
        # older style seed definition
        m = re.findall('vSeeds.push_back\(CDNSSeedData\("([\w\d.-]+)", "([-\w\d.-]+)"(, true|)\)\);', content)
        seed_index = 1

    for seed in m:
        if 'test' not in seed[seed_index]:
            data['seed_nodes'].append(seed[seed_index])

    return data


def test_regexes(content, *regexes):
    for regex in regexes:
        m = re.search(regex, content)
        if m:
            return m.groups()[0] if len(m.groups()) else True



from moneywagon.services import Service, BitcoinAbe, BitpayInsight

class BlockStrap(Service):
    """
    Documentation here: http://docs.blockstrap.com/en/api/
    """
    #decomissioned
    service_id = 10
    domain = 'http://api.blockstrap.com'
    supported_cryptos = ['btc', 'ltc', 'drk', 'doge']
    api_homepage = "http://blockstrap.com/blog/deprecated/"

    def get_balance(self, crypto, address, confirmations=None):
        url = "%s/v0/%s/address/id/%s" % (self.domain, crypto, address)
        response = self.get_url(url).json()
        return response['data']['address']['balance'] / 1e8

    def push_tx(self, crypto, tx_hex):
        url = "%s/v0/%s/transaction/relay/%s" % (self.domain, crypto, tx_hex)
        return self.get_url(url).json()['data']['id']

    def get_transactions(self, crypto, address):
        url = "%s/v0/%s/address/transactions/%s" % (self.domain, crypto, address)
        txs = []
        for tx in self.get_url(url).json()['data']['address']['transactions']:
            s_amount = tx['tx_address_input_value'] or tx['tx_address_output_value'] * -1
            txs.append(dict(
                date=arrow.get(tx['block_time']).datetime,
                amount=s_amount / 1e8,
                confirmations=tx['confirmations'],
                txid=tx['id'],
            ))
        return txs

    def get_unspent_outputs(self, crypto, address):
        url = "%s/v0/%s/address/unspents/%s" % (self.domain, crypto, address)
        utxos = []
        for utxo in self.get_url(url).json()['data']['address']['transactions']:
            utxos.append(dict(
                amount=utxo['tx_address_value'],
                address=address,
                output="%s:%s" % (utxo['id'].lower(), utxo['tx_address_tx_pos']),
                confirmations=utxo['confirmations']
            ))
        return utxos

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        if block_hash:
            url = "%s/v0/%s/block/id/%s" % (self.domain, crypto, block_hash)
            b = "block"
        elif block_number:
            url = "%s/v0/%s/block/height/%s" % (self.domain, crypto, block_number)
            b = "blocks"
        elif latest:
            url = "%s/v0/%s/block/latest" % (self.domain, crypto)
            b = 'block'

        r = self.get_url(url).json()['data'][b]

        if block_number:
            r = r[0]

        return dict(
            block_number=r['height'],
            confirmations=r["confirmations"],
            time=arrow.get(r['time_display']).datetime,
            sent_value=r["output_value"] / 1e8,
            total_fees=r["fees"] / 1e8,
            size=r['size'],
            hash=r['id'].lower(),
            merkle_root=r['merkel_root'].lower(),
            previous_hash=r['prev_block_id'].lower(),
            next_hash=r['next_block_id'].lower(),
            txids=sorted(x['id'] for x in r['transactions']),
            tx_count=len(r['transactions'])
        )

class LitecoinAbe(BitcoinAbe):
    service_id = 16
    supported_cryptos = ['ltc']
    base_url = "http://bitcoin-abe.info/chain/Litecoin"


class NamecoinAbe(BitcoinAbe):
    service_id = 17
    supported_cryptos = ['nmc']
    base_url = "http://bitcoin-abe.info/chain/Namecoin"

class CoinSwap(Service):
    service_id = 26
    def get_current_price(self, crypto, fiat):
        chunk = ("%s/%s" % (crypto, fiat)).upper()
        url = "https://api.coin-swap.net/market/stats/%s" % chunk
        response = self.get_url(url).json()
        return float(response['lastprice']), 'coin-swap'


class ExCoIn(Service):
    service_id = 27
    # decommissioned
    def get_current_price(self, crypto, fiat):
        url = "https://api.exco.in/v1/exchange/%s/%s" % (fiat, crypto)
        response = self.get_url(url).json()
        return float(response['last_price']), 'exco.in'

class TheBitInfo(BitpayInsight):
    service_id = 29
    supported_cryptos = ['btc']
    domain = "insight.thebit.info"
    name = "TheBit.info"

class FTCc(BitpayInsight):
    service_id = 34
    supported_cryptos = ['ftc']
    domain = "block.ftc-c.com"
    name = "Feathercoin China"

class FeathercoinCom(Service):
    service_id = 21
    supported_cryptos = ['ftc']
    api_homepage = "http://api.feathercoin.com/"
    name = "Feathercoin.com"

    def get_balance(self, crypto, address, confirmations=1):
        url= "http://api.feathercoin.com/?output=balance&address=%s&json=1" % address
        response = self.get_url(url)
        return float(response.json()['balance'])

class Toshi(Service):
    api_homepage = "https://toshi.io/docs/"
    service_id = 6
    url = "https://bitcoin.toshi.io/api/v0"
    name = "Toshi"

    supported_cryptos = ['btc']

    def check_error(self, response):
        if 'error' in response.json():
            raise ServiceError("Toshi returned error: %s" % response.json()['error'])

        if response.status_code == 404:
            return # don't skip on 404

        super(Toshi, self).check_error(response)

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/addresses/%s" % (self.url, address)
        response = self.get_url(url).json()
        return response['balance'] / 1e8

    def get_transactions(self, crypto, address, confirmations=1):
        url = "%s/addresses/%s/transactions" % (self.url, address)
        response = self.get_url(url)
        if response.status_code == 404:
            return []

        j = response.json()
        transactions = []
        for tx in j.get('unconfirmed_transactions', []) + j['transactions']:
            transactions.append(dict(
                amount=sum([x['amount'] / 1e8 for x in tx['outputs'] if address in x['addresses']]),
                txid=tx['hash'],
                date=arrow.get(tx['block_time']).datetime if tx.get('block_time', False) else None,
                confirmations=tx['confirmations']
            ))

        return transactions

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s/addresses/%s/unspent_outputs" % (self.url, address)
        response = self.get_url(url).json()
        utxos = []
        for utxo in response:
            cons = utxo['confirmations']
            if cons < confirmations:
                continue
            utxos.append(dict(
                amount=utxo['amount'],
                address=address,
                output="%s:%s" % (utxo['transaction_hash'], utxo['output_index']),
                confirmations=cons,
                vout=utxo['output_index'],
                txid=utxo['transaction_hash'],
                scriptPubKey=utxo['script_hex'],
                scriptPubKey_asm=utxo['script']
            ))
        return utxos


    def push_tx(self, crypto, tx_hex):
        url = "%s/transactions/%s" % (self.url, tx_hex)
        return self.get_url(url).json()['hash']

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        if latest:
            url = "%s/blocks/latest" % self.url
        else:
            url = "%s/blocks/%s%s" % (
                self.url, block_hash if block_hash != '' else '',
                block_number if block_number != '' else ''
            )

        r = self.get_url(url).json()
        return dict(
            block_number=r['height'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time']).datetime,
            sent_value=r['total_out'] / 1e8,
            total_fees=r['fees'] / 1e8,
            mining_difficulty=r['difficulty'],
            size=r['size'],
            hash=r['hash'],
            merkle_root=r['merkle_root'],
            previous_hash=r['previous_block_hash'],
            next_hash=r['next_blocks'][0]['hash'] if len(r['next_blocks']) else None,
            txids=sorted(r['transaction_hashes']),
            tx_count=len(r['transaction_hashes'])
        )

class NeoCrypto(BitpayInsight):
    service_id = 51
    protocol = "https"
    domain = "insight.neocrypto.io"
    api_tag = 'insight-api'
    name = "Neocrypt"

class VertcoinInfo(Iquidus):
    service_id = 56
    name = "Vertcoin.info"
    base_url = "http://explorer.vertcoin.info"
    supported_cryptos = ['vtc']

class Blockr(Service):
    service_id = 5
    supported_cryptos = ['btc', 'ltc', 'ppc', 'mec', 'qrk', 'dgc', 'tbtc']
    api_homepage = "http://blockr.io/documentation/api"

    explorer_address_url = "http://blockr.io/address/info/{address}"
    explorer_tx_url = "http://blockr.io/address/info/{txid}"
    explorer_blockhash_url = "http://blockr.io/block/info/{blockhash}"
    explorer_blocknum_url = "http://blockr.io/block/info/{blocknum}"
    explorer_latest_block = "http://blockr.io/block/info/latest"

    json_address_url = "http://{crypto}.blockr.io/api/v1/address/info/{address}"
    json_single_tx_url = "http://{crypto}.blockr.io/api/v1/tx/info/{txid}"
    json_txs_url = url = "http://{crypto}.blockr.io/api/v1/address/txs/{address}?unconfirmed=1"
    json_unspent_outputs_url = "http://{crypto}.blockr.io/api/v1/address/unspent/{address}"
    name = "Blockr.io"

    def get_balance(self, crypto, address, confirmations=1):
        url = self.json_address_url.format(address=address, crypto=crypto)
        response = self.get_url(url)
        return response.json()['data']['balance']

    def get_balance_multi(self, crypto, addresses, confirmations=1):
        url = self.json_address_url.format(address=','.join(addresses), crypto=crypto)
        balances = {}
        for bal in self.get_url(url).json()['data']:
            balances[bal['address']] = bal['balance']
        return balances

    def _format_tx(self, tx, address):
        return dict(
            date=arrow.get(tx['time_utc']).datetime,
            amount=tx['amount'],
            txid=tx['tx'],
            confirmations=tx['confirmations'],
            addresses=[address],
        )

    def get_transactions(self, crypto, address, confirmations=1):
        url = self.json_txs_url.format(address=address, crypto=crypto)
        response = self.get_url(url)

        transactions = []
        for tx in response.json()['data']['txs']:
            transactions.append(self._format_tx(tx, address))
        return transactions

    def get_transactions_multi(self, crypto, addresses, confirmation=1):
        url = self.json_txs_url.format(address=','.join(addresses), crypto=crypto)
        transactions = []
        for data in self.get_url(url).json()['data']:
            for tx in data['txs']:
                transactions.append(self._format_tx(tx, data['address']))
        return transactions

    def _format_single_tx(self, tx):
        ins = [
            {
                'address': x['address'],
                'amount': currency_to_protocol(x['amount']) * -1,
                'txid': x['vout_tx']
            } for x in tx['vins']
        ]
        outs = [
            {
                'address': x['address'],
                'amount': currency_to_protocol(x['amount']),
                'scriptPubKey': x['extras']['script'] if 'extras' in x else None
            } for x in tx['vouts']
        ]

        return dict(
            time=arrow.get(tx['time_utc']).datetime,
            block_number=tx['block'],
            inputs=ins,
            outputs=outs,
            txid=tx['tx'],
            total_in=sum(x['amount'] for x in ins),
            total_out=sum(x['amount'] for x in outs),
            confirmations=tx['confirmations'],
            fee=float(tx['fee'])
        )

    def get_single_transaction(self, crypto, txid):
        url = self.json_single_tx_url.format(crypto=crypto, txid=txid)
        r = self.get_url(url).json()['data']
        return self._format_single_tx(r)


    def get_single_transaction_multi(self, crypto, txids):
        url = self.json_single_tx_url.format(crypto=crypto, txid=','.join(txids))
        txs = []
        for tx in self.get_url(url).json()['data']:
            txs.append(self._format_single_tx(tx))
        return txs

    def _format_utxo(self, utxo, address):
        return dict(
            amount=currency_to_protocol(utxo['amount']),
            address=address,
            output="%s:%s" % (utxo['tx'], utxo['n']),
            txid=utxo['tx'],
            vout=utxo['n'],
            confirmations=utxo['confirmations'],
            scriptPubKey=utxo['script']
        )

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = self.json_unspent_outputs_url.format(address=address, crypto=crypto)
        utxos = []
        for utxo in self.get_url(url).json()['data']['unspent']:
            cons = utxo['confirmations']
            if cons < confirmations:
                continue
            utxos.append(self._format_utxo(utxo, address))
        return utxos

    def get_unspent_outputs_multi(self, crypto, addresses, confirmations=1):
        url = self.json_unspent_outputs_url.format(address=','.join(addresses), crypto=crypto)
        utxos = []
        for data in self.get_url(url).json()['data']:
            for utxo in data['unspent']:
                cons = utxo['confirmations']
                if cons < confirmations:
                    continue
                utxos.append(self._format_utxo(utxo, data['address']))
        return utxos

    def push_tx(self, crypto, tx_hex):
        url = "http://%s.blockr.io/api/v1/tx/push" % crypto
        resp = self.post_url(url, {'tx': tx_hex}).json()
        if resp['status'] == 'fail':
            raise ValueError(
                "Blockr returned error: %s %s %s" % (
                    resp['code'], resp['data'], resp['message']
                )
            )
        return resp['data']

    def get_block(self, crypto, block_hash=None, block_number=None, latest=False):
        if block_number == 0:
            raise SkipThisService("Block 0 not supported (bug in Blockr.io?)")

        url ="http://%s.blockr.io/api/v1/block/info/%s%s%s" % (
            crypto,
            block_hash if block_hash is not None else '',
            block_number if block_number is not None else '',
            'last' if latest else ''
        )
        r = self.get_url(url).json()['data']
        return dict(
            block_number=r['nb'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time_utc']).datetime,
            sent_value=r['vout_sum'],
            total_fees=float(r['fee']),
            mining_difficulty=r['difficulty'],
            size=int(r['size']),
            hash=r['hash'],
            merkle_root=r['merkleroot'],
            previous_hash=r['prev_block_hash'],
            next_hash=r['next_block_hash'],
            tx_count=r['nb_txs'],
        )

class ZCLexplorer(BitpayInsight):
    service_id = 126
    domain = "zclexplorer.org"
    protocol = "http"
    supported_cryptos = ['zcl']
    name = "Zclassic Insight"
    version = 0.0



from .blockchain_services import *
from .exchange_services import *

def get_service(name=None, id=None):
    from moneywagon import ALL_SERVICES
    for service in ALL_SERVICES:
        if (name and service.name.lower() == name.lower()) or (id and service.id == id):
            return service

def _get_all_services(crypto=None, just_exchange=False):
    """
    Go through the crypto_data structure and return all list of all (unique)
    installed services. Optionally filter by crypto-currency.
    """
    from moneywagon.crypto_data import crypto_data

    if not crypto:
        # no currency specified, get all services
        to_iterate = crypto_data.items()
    else:
        # limit to one currency
        to_iterate = [(crypto, crypto_data[crypto])]

    services = []
    for currency, data in to_iterate:
        if 'services' not in data:
            continue
        if currency == '':
            continue # template

        # price services are defined as dictionaries, all other services
        # are defined as a list.
        price_services = data['services']['current_price']
        del data['services']['current_price']

        all_services = list(price_services.values())
        if not just_exchange:
            all_services += list(data['services'].values())

        # replace
        data['services']['current_price'] = price_services

        services.append([
            item for sublist in all_services for item in sublist
        ])

    return sorted(
        set([item for sublist in services for item in sublist]),
        key=lambda x: x.__name__
    )

class BadService(Service):
    service_id = 0

    def get_balance(self, crypto, address, confirmations=1):
        class FakeResponse(object):
            def json(self):
                return {}
        self.last_raw_response = FakeResponse()
        return 999999



import json
import os

from requests.auth import HTTPBasicAuth
from moneywagon.core import (
    Service, NoService, NoData, ServiceError, SkipThisService, currency_to_protocol,
    decompile_scriptPubKey
)
from bitcoin import deserialize
import arrow

from bs4 import BeautifulSoup
import re

import hmac, hashlib, time, requests, base64
from requests.auth import AuthBase

try:
    from urllib import urlencode, quote_plus
except ImportError:
    from urllib.parse import urlencode, quote_plus

def make_standard_nonce(small=False):
    num = int(time.time() * 1000)
    if small:
        return str(num - 1506215312123)
    return str(num)

def make_stateful_nonce(exchange):
    path = os.path.expanduser('~/.moneywagon_state')
    if not os.path.exists(path):
        # make
        with open(path, "w+") as f:
            f.write('{}')

    with open(path) as f:
        j = json.loads(f.read())
        if exchange not in j:
            j[exchange] = {'last_used_nonce': 0}

        nonce = j[exchange].get('last_used_nonce', 0) + 1
        j[exchange]['last_used_nonce'] = nonce

    with open(path, "w") as f:
        f.write(json.dumps(j))

    return nonce

def eight_decimal_places(amount, format="str"):
    """
    >>> eight_decimal_places(3.12345678912345)
    "3.12345679"
    >>> eight_decimal_places("3.12345678912345")
    "3.12345679"
    >>> eight_decimal_places(3.12345678912345, format='float')
    3.12345679
    >>> eight_decimal_places("3.12345678912345", format='float')
    3.12345679
    """
    if type(amount) == str:
        return amount
    if format == 'str':
        return "%.8f" % amount
    if format == 'float':
        return float("%.8f" % amount)

class Bitstamp(Service):
    service_id = 1
    supported_cryptos = ['btc']
    api_homepage = "https://www.bitstamp.net/api/"
    name = "Bitstamp"
    exchange_fee_rate = 0.0025

    def __init__(self, customer_id=None, **kwargs):
        self.customer_id = customer_id
        super(Bitstamp, self).__init__(**kwargs)

    def make_market(self, crypto, fiat):
        return ("%s%s" % (crypto, fiat)).lower()

    def get_current_price(self, crypto, fiat):
        url = "https://www.bitstamp.net/api/v2/ticker/%s" % (
            self.make_market(crypto, fiat)
        )
        response = self.get_url(url).json()
        return float(response['last'])

    def get_pairs(self):
        return ['btc-usd', 'btc-eur', 'bch-btc', 'bch-usd', 'xrp-usd', 'xrp-eur', 'xrp-btc']

    def get_orderbook(self, crypto, fiat):
        url = "https://www.bitstamp.net/api/v2/order_book/%s/" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'bids': [(float(x[0]), float(x[1])) for x in resp['bids']],
            'asks': [(float(x[0]), float(x[1])) for x in resp['asks']]
        }

    def _make_signature(self, nonce):
        message = nonce + self.customer_id + self.api_key
        return hmac.new(
            self.api_secret,
            msg=message,
            digestmod=hashlib.sha256
        ).hexdigest().upper()

    def _auth_request(self, url, params):
        nonce = make_standard_nonce()
        params.update({
            'nonce': nonce,
            'signature': self._make_signature(nonce),
            'key': self.api_key,
        })
        return self.post_url(url, params)

    def get_exchange_balance(self, currency, type="available"):
        url = "https://www.bitstamp.net/api/balance/"
        resp = self._auth_request(url, {}).json()
        try:
            return float(resp["%s_%s" % (currency.lower(), type)])
        except KeyError:
            return 0

    def get_total_exchange_balances(self):
        url = "https://www.bitstamp.net/api/balance/"
        resp = self._auth_request(url, {}).json()
        return {
            code[:-8]: float(bal) for code, bal in resp.items()
            if code.endswith("balance") and float(bal) > 0
        }

    def get_deposit_address(self, currency):
        if currency.lower() == 'btc':
            url = "https://www.bitstamp.net/api/bitcoin_deposit_address/"
            return self._auth_request(url, {}).json()
        if currency.lower() == 'xrp':
            url = "https://www.bitstamp.net/api/ripple_address/"
            return self._auth_request(url, {}).json()['address']
        if currency.lower() in ['eth', 'ltc']:
            url = "https://www.bitstamp.net/api/v2/%s_address/" % currency.lower()
            return self._auth_request(url, {}).json()['address']

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        if type == 'limit':
            url = "https://www.bitstamp.net/api/v2/%s/%s/" % (
                side, self.make_market(crypto, fiat)
            )
            resp = self._auth_request(url, {
                'amount': eight_decimal_places(amount),
                'price': price,
            })
        if type == 'market':
            url = "https://www.bitstamp.net/api/v2/%s/market/%s/" % (
                side, self.make_market(crypto, fiat)
            )
            resp = self._auth_request(url, {
                'amount': eight_decimal_places(amount),
            })

        return resp.json()
    make_order.supported_types = ['limit', 'market']
    make_order.minimums = {}


class CoinbaseExchangeAuth(AuthBase):
    def __init__(self, api_key, secret_key, passphrase):
        self.api_key = api_key
        self.secret_key = secret_key
        self.passphrase = passphrase

    def __call__(self, request):
        timestamp = str(time.time())
        message = timestamp + request.method + request.path_url + (request.body or '')
        hmac_key = base64.b64decode(self.secret_key)
        signature = hmac.new(hmac_key, message, hashlib.sha256)
        signature_b64 = signature.digest().encode('base64').rstrip('\n')

        request.headers.update({
            'CB-ACCESS-SIGN': signature_b64,
            'CB-ACCESS-TIMESTAMP': timestamp,
            'CB-ACCESS-KEY': self.api_key,
            'CB-ACCESS-PASSPHRASE': self.passphrase,
            'Content-Type': 'application/json'
        })
        return request

class GDAX(Service):
    service_id = 59
    name = "GDAX"
    base_url = "https://api.gdax.com"
    api_homepage = "https://docs.gdax.com/"
    supported_cryptos = ['btc', 'ltc', 'eth', 'bch']
    exchange_fee_rate = 0.0025

    def __init__(self, api_pass=None, **kwargs):
        self.auth = None
        self.api_pass = api_pass

        super(GDAX, self).__init__(**kwargs)

        if self.api_key and self.api_secret and self.api_pass:
            self.auth = CoinbaseExchangeAuth(self.api_key, self.api_secret, self.api_pass)

    def check_error(self, response):
        if response.status_code != 200:
            j = response.json()
            raise ServiceError("GDAX returned %s error: %s" % (
                response.status_code, j['message'])
            )

        super(GDAX, self).check_error(response)

    def make_market(self, crypto, fiat):
        return ("%s-%s" % (crypto, fiat)).upper()

    def get_current_price(self, crypto, fiat):
        url = "%s/products/%s/ticker" % (self.base_url, self.make_market(crypto, fiat))
        response = self.get_url(url).json()
        return float(response['price'])

    def get_pairs(self):
        url = "%s/products" % self.base_url
        r = self.get_url(url).json()
        return [x['id'].lower() for x in r]

    def get_orderbook(self, crypto, fiat):
        url = "%s/products/%s/book?level=3" % (self.base_url, self.make_market(crypto, fiat))
        r = self.get_url(url).json()
        return {
            'bids': [(float(x[0]), float(x[1])) for x in r['bids']],
            'asks': [(float(x[0]), float(x[1])) for x in r['asks']]
        }

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        time_in_force = 'GTC'
        if type == 'fill-or-kill':
            type = 'limit'
            time_in_force = 'FOK'

        url = "%s/orders" % self.base_url
        data = {
            "size": eight_decimal_places(amount),
            "type": type,
            "price": price,
            "side": side,
            "time_in_force": time_in_force,
            "product_id": self.make_market(crypto, fiat)
        }
        response = self.post_url(url, json=data, auth=self.auth).json()
        return response['id']
    make_order.supported_types = ['fill-or-kill', 'limit', 'market', 'stop']
    make_order.minimums = {'btc': 0.0001, 'eth': 0.001, 'ltc': 0.01, 'usd': 1, 'eur': 1, 'gbp': 1}

    def list_orders(self, status="open"):
        url = "%s/orders" % self.base_url
        response = self.get_url(url, auth=self.auth)
        return response.json()

    def cancel_order(self, order_id):
        url = "%s/orders/%s" % (self.base_url, order_id)
        response = self.delete_url(url, auth=self.auth)
        return response.json()

    def get_exchange_balance(self, currency, type="available"):
        url = "%s/accounts" % self.base_url
        resp = self.get_url(url, auth=self.auth).json()
        try:
            match = [x for x in resp if currency.upper() == x['currency']][0]
        except IndexError:
            return 0
        return float(match[type])

    def get_total_exchange_balances(self):
        url = "%s/accounts" % self.base_url
        resp = self.get_url(url, auth=self.auth).json()
        return {
            x['currency'].lower(): float(x['balance']) for x in resp
        }

    def initiate_withdraw(self, currency, amount, address):
        url = "%s/withdrawals/crypto" % self.base_url
        resp = self.post_url(url, auth=self.auth, json={
            'crypto_address': address,
            'currency': currency.upper(),
            'amount': eight_decimal_places(amount)
        })
        return resp.json()

class BitFinex(Service):
    service_id = 120
    api_homepage = "https://bitfinex.readme.io/v2/reference"
    exchange_fee_rate = 0.002

    def check_error(self, response):
        j = response.json()
        if j and type(j) is list and j[0] == 'error':
            raise SkipThisService(
                "BitFinex returned Error: %s (%s)" % (j[2], j[1])
            )
        super(BitFinex, self).check_error(response)

    def parse_market(self, market):
        crypto = market[:3]
        fiat = market[3:]
        if crypto == 'dsh':
            crypto = 'dash'
        if crypto == 'iot':
            crypto = 'miota'
        return crypto, fiat

    def fix_symbol(self, symbol):
        if symbol == 'dash':
            return 'dsh'
        if symbol == 'miota':
            return 'iot'
        return symbol

    def make_market(self, crypto, fiat):
        return "%s%s" % (
            self.fix_symbol(crypto).lower(), self.fix_symbol(fiat).lower()
        )

    def get_pairs(self):
        url = "https://api.bitfinex.com/v1/symbols"
        r = self.get_url(url).json()
        return ["%s-%s" % self.parse_market(x) for x in r]

    def get_current_price(self, crypto, fiat):
        url = "https://api.bitfinex.com/v2/ticker/t%s" % self.make_market(crypto, fiat).upper()
        r = self.get_url(url).json()
        return r[6]

    def get_orderbook(self, crypto, fiat):
        url = "https://api.bitfinex.com/v1/book/%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'bids': [(float(x['price']), float(x['amount'])) for x in resp['bids']],
            'asks': [(float(x['price']), float(x['amount'])) for x in resp['asks']]
        }

    def _make_signature(self, path, args, nonce, version=2):
        if version == 2:
            msg = '/api/' + path + nonce + json.dumps(args)
        elif version == 1:
            msg = nonce # actually payload, but passed in as nonce
        return hmac.new(self.api_secret, msg, hashlib.sha384).hexdigest()

    def _auth_request(self, path, params):
        url = "https://api.bitfinex.com"
        nonce = make_standard_nonce()
        if path.startswith("/v2/"):
            headers = {
                'bfx-nonce': nonce,
                'bfx-apikey': self.api_key,
                'bfx-signature': self._make_signature(path, params, nonce, version=2)
            }
        elif path.startswith("/v1/"):
            params['request'] = path
            params['nonce'] = nonce
            payload = base64.b64encode(json.dumps(params))
            headers = {
                'X-BFX-PAYLOAD': payload,
                'X-BFX-APIKEY': self.api_key,
                'X-BFX-SIGNATURE': self._make_signature(path, params, payload, version=1)
            }
        return self.post_url(url + path, json=params, headers=headers)

    def get_deposit_address(self, crypto):
        resp = self._auth_request("/v2/auth/r/wallets", {})
        filt = [x[2] for x in resp.json() if x[1] == crypto.upper()]
        return filt[0] if filt else 0

    def get_exchange_balance(self, currency, type="available"):
        curr = self.fix_symbol(currency)
        resp = self._auth_request("/v1/balances", {}).json()
        for item in resp:
            if item['currency'] == curr.lower():
                return float(item[type])
        return 0

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        url = "/v1/order/new"
        resp = self._auth_request(url, {
            'symbol': self.make_market(crypto, fiat),
            'amount': eight_decimal_places(amount),
            'price': str(price),
            'side': side,
            'type': 'exchange %s' % type,
        })
        return resp.json()
    make_order.supported_types = ['fill-or-kill', 'market', 'limit', 'stop', 'trailing-stop']
    make_order.minimums = {}

    def initiate_withdraw(self, crypto, amount, address):
        from moneywagon.crypto_data import crypto_data
        if crypto == 'etc':
            type = "ethereumc"
        else:
            type = crypto_data[crypto.lower()]['name'].lower()
        resp = self._auth_request("/v1/withdraw", {
            "withdraw_type": type,
            "walletselected": "exchange",
            "amount": eight_decimal_places(amount),
            "address": address,
        }).json()
        return resp


class NovaExchange(Service):
    service_id = 89
    name = "NovaExchange"
    api_homepage = "https://novaexchange.com/remote/faq/"

    def make_market(self, crypto, fiat):
        return "%s_%s" % (fiat, crypto)

    def check_error(self, response):
        if response.json()['status'] == 'error':
            raise ServiceError("NovaExchange returned error: %s" % response.json()['message'])

        super(NovaExchange, self).check_error(response)

    def get_current_price(self, crypto, fiat):
        url = "https://novaexchange.com/remote/v2/market/info/%s" % self.make_market(crypto, fiat)
        r = self.get_url(url).json()
        return float(r['markets'][0]['last_price'])

    def get_pairs(self):
        url = "https://novaexchange.com/remote/v2/markets/"
        r = self.get_url(url).json()
        ret = []
        for pair in r['markets']:
            fiat = pair['basecurrency'].lower()
            crypto = pair['currency'].lower()
            ret.append("%s-%s" % (crypto, fiat))
        return ret

    def get_orderbook(self, crypto, fiat):
        url = "https://novaexchange.com/remote/v2/market/openorders/%s/both/" % (
            self.make_market(crypto, fiat)
        )
        r = self.get_url(url).json()
        return {
            'bids': [(float(x['price']), float(x['amount'])) for x in r['buyorders']],
            'asks': [(float(x['price']), float(x['amount'])) for x in r['sellorders']],
        }

    def _make_signature(self, url):
        return base64.b64encode(
            hmac.new(self.api_secret, url, hashlib.sha512).digest()
        )

    def _auth_request(self, url, params):
        url += '?nonce=' + make_standard_nonce()
        params['apikey'] = self.api_key
        params['signature'] = self._make_signature(url)
        headers = {'content-type': 'application/x-www-form-urlencoded'}
        return self.post_url(url, data=params, headers=headers, timeout=60)

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        url = "https://novaexchange.com/remote/v2/private/trade/%s/" % (
            self.make_market(crypto, fiat)
        )
        params = {
            'tradetype': side.upper(),
            'tradeamount': eight_decimal_places(amount),
            'tradeprice': price,
            'tradebase': 0, # indicates "amount" is in crypto units, not fiat units
        }
        resp = self._auth_request(url, params)
        return resp.json()['tradeitems'][0]['orderid']
    make_order.minimums = {}

    def cancel_order(self, order_id):
        url = "https://novaexchange.com/remote/v2/private/cancelorder/%s/" % order_id
        resp = self._auth_request(url, {})
        return resp.json()['status'] == 'ok'

    def list_orders(self, status="open"):
        if status == 'open':
            url = "https://novaexchange.com/remote/v2/private/myopenorders/"
        else:
            NotImplementedError("getting orders by status=%s not implemented yet" % status)
        resp = self._auth_request(url, {})
        return resp.json()['items']

    def get_deposit_address(self, crypto):
        url = "https://novaexchange.com/remote/v2/private/getdepositaddress/%s/" % crypto
        resp = self._auth_request(url, {})
        return resp.json()['address']

    def initiate_withdraw(self, crypto, amount, address):
        url = "https://novaexchange.com/remote/v2/private/withdraw/%s/" % crypto
        params = {'currency': crypto, 'amount': eight_decimal_places(amount), 'address': address}
        resp = self._auth_request(url, params)
        return resp.json()


class xBTCe(Service):
    service_id = 90
    name = "xBTCe"
    api_homepage = "https://www.xbtce.com/tradeapi"

    def get_current_price(self, crypto, fiat):
        if crypto.lower() == 'dash':
            crypto = "dsh"
        if fiat.lower() == 'rur':
            fiat = 'rub'
        if fiat.lower() == 'cny':
            fiat = 'cnh'
        pair = "%s%s" % (crypto.upper(), fiat.upper())
        url = "https://cryptottlivewebapi.xbtce.net:8443/api/v1/public/ticker/%s" % pair
        r = self.get_url(url).json()
        try:
            return r[0]['LastSellPrice']
        except IndexError:
            raise ServiceError("Pair not found")

    def get_pairs(self):
        url = "https://cryptottlivewebapi.xbtce.net:8443/api/v1/public/symbol"
        r = self.get_url(url).json()
        ret = []
        for pair in r:
            crypto = pair['MarginCurrency'].lower()
            fiat = pair['ProfitCurrency'].lower()

            if crypto.lower() == 'dsh':
                crypto = "dash"
            if fiat.lower() == 'rub':
                fiat = 'rur'
            if fiat == 'cnh':
                fiat = 'cny'
            ret.append(("%s-%s" % (crypto, fiat)))

        return list(set(ret))


class OKcoin(Service):
    service_id = 60
    name = "OKcoin"
    exchange_base_url = "https://www.okcoin.cn"
    block_base_url = "http://block.okcoin.cn"
    supported_cryptos = ['btc', 'ltc']
    api_homepage = "https://www.okcoin.cn/rest_getStarted.html"

    def get_current_price(self, crypto, fiat):
        if not fiat == 'cny':
            raise SkipThisService("Only fiat=CNY supported")

        url = "%s/api/v1/ticker.do?symbol=%s_%s" % (
            self.exchange_base_url, crypto.lower(), fiat.lower()
        )
        response = self.get_url(url).json()
        return float(response['ticker']['last'])

    def check_error(self, response):
        j = response.json()
        if 'error_code' in j:
            raise ServiceError("OKcoin returned error code %s" % j['error_code'])

        super(OKcoin, self).check_error(response)

    def get_pairs(self):
        return ['btc-cny', 'ltc-cny']

    def get_block(self, crypto, block_hash=None, block_number=None, latest=False):
        if latest:
            args = 'latest_block.do?'

        if block_number or block_number == 0:
            args = "block_height.do?block_height=%s&" % block_number

        if block_hash:
            raise SkipThisService("Block by hash not supported")

        url = "%s/api/v1/%ssymbol=%s" % (
            self.block_base_url, args, crypto.upper()
        )
        r = self.get_url(url).json()

        ret = dict(
            block_number=r['height'],
            size=r['size'],
            time=arrow.get(r['time'] / 1000).datetime,
            hash=r['hash'],
            txids=r['txid'],
            tx_count=r['txcount'],
            version=r['version'],
            mining_difficulty=r['difficulty'],
            total_fees=r['fee'],
            sent_value=r['totalOut']
        )

        if r.get('relayed_by'):
            ret['miner'] = r['relayed_by']

        if r.get('previousblockhash'):
            ret['previous_hash'] = r['previousblockhash']

        if r.get('nextblockhash'):
            ret['next_hash'] = r['nextblockhash']

        return ret

class FreeCurrencyConverter(Service):
    service_id = 61
    base_url = "http://free.currencyconverterapi.com"
    api_homepage = "http://www.currencyconverterapi.com/docs"

    def get_fiat_exchange_rate(self, from_fiat, to_fiat):
        pair = "%s_%s" % (to_fiat.upper(), from_fiat.upper())
        url = "%s/api/v3/convert?q=%s&compact=y" % (
            self.base_url, pair
        )
        response = self.get_url(url).json()
        return response[pair]['val']


class BTCChina(Service):
    service_id = 62
    api_homepage = "https://www.btcc.com/apidocs/spot-exchange-market-data-rest-api#ticker"
    name = "BTCChina"

    def get_current_price(self, crypto, fiat):
        if fiat == 'usd':
            url = "https://spotusd-data.btcc.com/data/pro/ticker?symbol=%sUSD" % crypto.upper()
            key = "Last"
        else:
            url = "https://data.btcchina.com/data/ticker?market=%s%s" % (
                crypto.lower(), fiat.lower()
            )
            key = "last"
        response = self.get_url(url).json()
        if not response:
            raise ServiceError("Pair not supported (blank response)")
        return float(response['ticker'][key])


class Gemini(Service):
    service_id = 63
    api_homepage = "https://docs.gemini.com/rest-api/"
    name = "Gemini"
    exchange_fee_rate = 0.0025

    def check_error(self, response):
        j = response.json()
        if 'result' in j and j['result'] == 'error':
            raise ServiceError("Gemini returned error: %s" % j['reason'])

        super(Gemini, self).check_error(response)

    def make_market(self, crypto, fiat):
        return "%s%s" % (crypto.lower(), fiat.lower())

    def get_current_price(self, crypto, fiat):
        url = "https://api.gemini.com/v1/pubticker/%s" % self.make_market(crypto, fiat)
        response = self.get_url(url).json()
        return float(response['last'])

    def get_orderbook(self, crypto, fiat):
        url = "https://api.gemini.com/v1/book/%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'bids': [(float(x['price']), float(x['amount'])) for x in resp['bids']],
            'asks': [(float(x['price']), float(x['amount'])) for x in resp['asks']]
        }

    def _make_signature(self, payload):
        return hmac.new(self.api_secret, payload, hashlib.sha384).hexdigest()

    def _auth_request(self, path, params):
        params['nonce'] = make_standard_nonce()
        params['request'] = path
        payload = base64.b64encode(json.dumps(params))
        headers = {
            'X-GEMINI-APIKEY': self.api_key,
            'X-GEMINI-PAYLOAD': payload,
            'X-GEMINI-SIGNATURE': self._make_signature(payload)
        }
        return self.post_url("https://api.gemini.com" + path, headers=headers)

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        path = "/v1/order/new"
        #order_token = "" % datetime.datetime.now()
        opts = []

        if type == 'fill-or-kill':
            opts = ['immediate-or-cancel']
            type = "limit"

        if type == 'post-only':
            opts = ["maker-or-cancel"]
            type = 'limit'

        if type != "limit":
            raise NotImplementedError("Only limit orders currently supported")

        params = {
            #"client_order_id": order_token, # A client-specified order token
            "symbol": self.make_market(crypto, fiat), # Or any symbol from the /symbols api
            "amount": eight_decimal_places(amount), # Once again, a quoted number
            "price": str(price),
            "side": side, # must be "buy" or "sell"
            "type": "exchange %s" % type, # the order type; only "exchange limit" supported
            "options": opts # execution options; may be omitted for a standard limit order
        }
        resp = self._auth_request(path, params).json()
        return resp
    make_order.supported_types = ['post-only', 'limit', 'fill-or-kill']
    make_order.minimums = {}

    def get_deposit_address(self, currency):
        path = "/v1/deposit/%s/newAddress" % currency.lower()
        resp = self._auth_request(path, {})
        return resp.json()['address']

    def get_exchange_balance(self, currency, type="available"):
        path = "/v1/balances"
        resp = self._auth_request(path, {})
        try:
            match = [x for x in resp.json() if x['currency'] == currency.upper()][0]
        except IndexError:
            return 0
        return float(match[type])

    def get_total_exchange_balances(self):
        path = "/v1/balances"
        resp = self._auth_request(path, {})
        return {
            x['currency'].lower(): float(x['amount']) for x in resp.json()
            if float(x['amount']) > 0
        }


class CexIO(Service):
    service_id = 64
    api_homepage = "https://cex.io/rest-api"
    name = "Cex.io"
    exchange_fee_rate = 0.002

    def __init__(self, user_id=None, **kwargs):
        self.user_id = user_id
        super(CexIO, self).__init__(**kwargs)

    def check_error(self, response):
        super(CexIO, self).check_error(response)
        j = response.json()
        if 'error' in j:
            raise ServiceError("CexIO returned error: %s" % j['error'])

    def get_current_price(self, crypto, fiat):
        url = "https://cex.io/api/ticker/%s/%s" % (crypto.upper(), fiat.upper())
        response = self.get_url(url).json()
        return float(response['last'])

    def get_pairs(self):
        url = "https://cex.io/api/currency_limits"
        r = self.get_url(url).json()['data']['pairs']
        return [("%s-%s" % (x['symbol1'], x['symbol2'])).lower() for x in r]

    def get_orderbook(self, crypto, fiat):
        url = "https://cex.io/api/order_book/%s/%s/" % (crypto.upper(), fiat.upper())
        resp = self.get_url(url).json()
        return {
            'bids': [(x[0], x[1]) for x in resp['bids']],
            'asks': [(x[0], x[1]) for x in resp['asks']]
        }

    def _make_signature(self, nonce):
        message = nonce + self.user_id + self.api_key
        return hmac.new(self.api_secret, message, hashlib.sha256).hexdigest().upper()

    def _auth_request(self, url, params):
        nonce = make_standard_nonce()
        params['nonce'] = nonce
        params['signature'] = self._make_signature(nonce)
        params['key'] = self.api_key
        return self.post_url(url, params)

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        url = "https://cex.io/api/place_order/%s/%s" % (crypto.upper(), fiat.upper())
        if type in ('limit', 'market'):
            print("about to send amount to cex:", eight_decimal_places(amount))
            params = {
                'type': side,
                'amount': eight_decimal_places(amount),
            }
            if type == 'market':
                params['order_type'] = 'market'
            if type == 'limit':
                params['price'] = price
        else:
            raise Exception("Order with type=%s not yet supported" % type)

        resp = self._auth_request(url, params).json()
        return resp['id']
    make_order.supported_types = ['limit', 'market']
    make_order.minimums = {'btc': 0.01}

    def list_orders(self):
        url = "https://cex.io/api/open_orders/"
        resp = self._auth_request(url, {})
        return resp.json()

    def cancel_order(self, order_id):
        url = "https://cex.io/api/cancel_order/"
        resp = self._auth_request(url, {'id': order_id})
        return resp.content == 'true'

    def get_deposit_address(self, crypto):
        url = "https://cex.io/api/get_address"
        resp = self._auth_request(url, {'currency': crypto.upper()})
        return resp.json()['data']

    def get_exchange_balance(self, currency, type="available"):
        url = "https://cex.io/api/balance/"
        resp = self._auth_request(url, {})
        try:
            return float(resp.json()[currency.upper()]['available'])
        except KeyError:
            return 0

    def get_total_exchange_balances(self):
        url = "https://cex.io/api/balance/"
        resp = self._auth_request(url, {})
        return {
            code.lower(): float(data['available']) for code, data in resp.json().items()
            if code not in ['timestamp', 'username'] and float(data['available']) > 0
        }

class Poloniex(Service):
    service_id = 65
    api_homepage = "https://poloniex.com/support/api/"
    name = "Poloniex"
    exchange_fee_rate = 0.0025

    def check_error(self, response):
        j = response.json()
        if 'error' in j:
            raise ServiceError("Poloniex returned error: %s" % j['error'])

        super(Poloniex, self).check_error(response)

    def fix_symbol(self, symbol):
        symbol = symbol.lower()
        if symbol == 'usd':
            return 'usdt'
        if symbol == 'xlm':
            return 'str'
        return symbol

    def reverse_fix_symbol(self, symbol):
        symbol = symbol.lower()
        if symbol == 'usdt':
            return 'usd'
        if symbol == 'str':
            return 'xlm'
        return symbol

    def get_current_price(self, crypto, fiat):
        url = "https://poloniex.com/public?command=returnTicker"
        response = self.get_url(url).json()

        is_usd = False
        if fiat.lower() == 'usd':
            fiat = 'usdt'
            is_usd = True

        find_pair = "%s_%s" % (fiat.upper(), crypto.upper())
        for pair, data in response.items():
            if pair == find_pair:
                return float(data['last'])

        reverse_pair = "%s_%s" % (crypto.upper(), fiat.upper())
        for pair, data in response.items():
            if pair == reverse_pair:
                return 1 / float(data['last'])

        btc_pair = "BTC_%s" % crypto.upper()
        if is_usd and btc_pair in response:
            btc_rate = float(response['USDT_BTC']['last'])
            fiat_exchange = float(response[btc_pair]['last'])
            return fiat_exchange * btc_rate

        raise SkipThisService("Pair %s not supported" % find_pair)

    def get_pairs(self):
        url = "https://poloniex.com/public?command=returnTicker"
        r = self.get_url(url).json()
        ret = []
        for pair in r.keys():
            fiat, crypto = pair.lower().split('_')
            ret.append("%s-%s" % (self.reverse_fix_symbol(crypto), self.reverse_fix_symbol(fiat)))
        return ret

    def get_orderbook(self, crypto, fiat):
        url = "https://poloniex.com/public?command=returnOrderBook&currencyPair=%s" % (
            self.make_market(crypto, fiat)
        )
        resp = self.get_url(url).json()
        return {
            'asks': [(float(x[0]), x[1]) for x in resp['asks']],
            'bids': [(float(x[0]), x[1]) for x in resp['bids']]
        }

    def make_market(self, crypto, fiat):
        return ("%s_%s" % (self.fix_symbol(fiat), self.fix_symbol(crypto))).upper()

    def _make_signature(self, args):
        str_args = urlencode(args)
        return hmac.new(self.api_secret, str_args, hashlib.sha512).hexdigest()

    def _auth_request(self, args):
        url = "https://poloniex.com/tradingApi"
        args["nonce"] = make_standard_nonce()
        headers = {
            'Sign': self._make_signature(args),
            'Key': self.api_key
        }
        return self.post_url(url, args, headers=headers)

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        params = {}
        if type == "fill-or-kill":
            params = {'fillOrKill': 1}
        if type == 'post-only':
            params = {'postOnly': 1}
        params.update({
            "command": side,
            "currencyPair": self.make_market(crypto, fiat),
            "rate": price,
            "amount": eight_decimal_places(amount)
        })
        r = self._auth_request(params)
        return r.json()['orderNumber']
    make_order.supported_types = ['limit', 'fill-or-kill', 'post-only']
    make_order.minimums = {}

    def cancel_order(self, order_id):
        r = self._auth_request({
            "command": "cancelOrder",
            "orderNumber": order_id
        })
        return r['success'] == 1

    def list_orders(self, crypto=None, fiat=None):
        if not crypto and not fiat:
            pair = "all"
        else:
            self.make_market(crypto, fiat)

        resp = self._auth_request({
            "command": "returnOpenOrders",
            "currencyPair": pair,
        })
        return resp.json()

    def initiate_withdraw(self, crypto, amount, address):
        resp = self._auth_request({
            "command": "withdrawl",
            "currency": crypto,
            "amount": eight_decimal_places(amount),
            "address": address
        })
        return resp.json()

    def get_deposit_address(self, currency):
        c = self.fix_symbol(currency)
        resp = self._auth_request({"command": "returnDepositAddresses"})
        address = resp.json().get(c.upper())
        if not address:
            return self.generate_new_deposit_address(c)
        return address

    def generate_new_deposit_address(self, crypto):
        resp = self._auth_request({
            "command": "generateNewAddress",
            "currency": crypto.upper()
        })
        return resp.json()['response']

    def get_exchange_balance(self, currency, type="available"):
        resp = self._auth_request({"command": "returnBalances"})
        return float(resp.json().get(self.reverse_fix_symbol(currency).upper()))

    def get_total_exchange_balances(self):
        resp = self._auth_request({"command": "returnBalances"})
        return {
            self.reverse_fix_symbol(code): float(bal) for code, bal in resp.json().items()
            if float(bal) > 0
        }

class Bittrex(Service):
    service_id = 66
    api_homepage = "https://bittrex.com/home/api"
    exchange_fee_rate = 0.0025

    def check_error(self, response):
        j = response.json()
        if not j['success']:
            raise ServiceError("Bittrex returned error: %s" % j['message'])

        super(Bittrex, self).check_error(response)

    def fix_symbol(self, symbol):
        if symbol.lower() == 'usd':
            return 'usdt'

        if symbol == 'xmy':
            return 'myr'

        if symbol == 'bcc':
            raise SkipThisService("BCC not supported (maybe you want BCH?)")

        if symbol == 'bch':
            return 'bcc'

        return symbol.lower()

    def reverse_fix_symbol(self, symbol):
        symbol = symbol.lower()
        if symbol == 'usdt':
            return 'usd'
        if symbol == 'bcc':
            return 'bch'
        return symbol

    def make_market(self, crypto, fiat):
        return "%s-%s" % (
            self.fix_symbol(fiat).upper(),
            self.fix_symbol(crypto).upper()
        )

    def get_current_price(self, crypto, fiat):
        url = "https://bittrex.com/api/v1.1/public/getticker?market=%s" % (
            self.make_market(crypto, fiat)
        )
        r = self.get_url(url).json()
        return r['result']['Last']

    def get_orderbook(self, crypto, fiat):
        url = "https://bittrex.com/api/v1.1/public/getorderbook?market=%s&type=both" % (
            self.make_market(crypto, fiat)
        )
        r = self.get_url(url).json()['result']
        return {
            'bids': [(x['Rate'], x['Quantity']) for x in r['buy']],
            'asks': [(x['Rate'], x['Quantity']) for x in r['sell']],
        }

    def get_pairs(self):
        url = "https://bittrex.com/api/v1.1/public/getmarkets"
        r = self.get_url(url).json()['result']
        ret = []
        for x in r:
            crypto = x['MarketCurrency'].lower()
            fiat = x['BaseCurrency'].lower()
            if fiat == 'usdt':
                fiat = 'usd'
            ret.append("%s-%s" % (crypto, fiat))
        return ret

    def _make_signature(self, url):
        return hmac.new(
            self.api_secret.encode(), url.encode(), hashlib.sha512
        ).hexdigest()

    def _auth_request(self, path, params):
        if not self.api_key or not self.api_secret:
            raise Exception("Trade API requires an API key and secret.")
        params["apikey"] = self.api_key
        params["nonce"] = make_standard_nonce()
        url = "https://bittrex.com/api" + path + "?" + urlencode(params)
        return self.get_url(url, headers={"apisign": self._make_signature(url)})

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        path = "/v1.1/market/%slimit" % side
        r = self._auth_request(path, {
            'market': self.make_market(crypto, fiat),
            'quantity': eight_decimal_places(amount),
            'rate': price
        })
        return r.json()['result']['uuid']
    make_order.supported_types = ['limit']
    make_order.minimums = {}

    def cancel_order(self, order_id):
        path = "/v1.1/market/cancel"
        r = self._auth_request(path, {'uuid': order_id})
        return r['success']

    def get_exchange_balance(self, currency, type="available"):
        currency = self.fix_symbol(currency)
        path = "/v1.1/account/getbalance"
        resp = self._auth_request(path, {'currency': self.fix_symbol(currency)}).json()['result']
        return resp[type.capitalize()] or 0

    def get_total_exchange_balances(self):
        path = "/v1.1/account/getbalances"
        resp = self._auth_request(path, {}).json()['result']
        return {
            self.reverse_fix_symbol(x['Currency']): x['Balance'] for x in resp
            if x['Balance'] > 0
        }


    def get_deposit_address(self, crypto):
        path = "/v1.1/account/getdepositaddress"
        resp = self._auth_request(path, {'currency': self.fix_symbol(crypto)})
        return resp.json()['result']['Address']

    def initiate_withdraw(self, crypto, amount, address):
        path = "/v1.1/account/withdraw"
        resp = self._auth_request(path, {
            'currency': self.fix_symbol(crypto),
            'quantity': eight_decimal_places(amount),
            'address': address
        })
        return resp.json()


class Huobi(Service):
    service_id = 67
    api_homepage = "https://github.com/huobiapi/API_Docs_en/wiki"
    name = "Huobi"

    def check_error(self, response):
        if response.status_code != 200:
            j = response.json()
            raise ServiceError("Huobi returned error: %s" % j['error'])

        super(Huobi, self).check_error(response)

    def get_current_price(self, crypto, fiat):
        if fiat.lower() == "cny":
            fiat = 'static'
        elif fiat.lower() == 'usd':
            pass
        else:
            raise SkipThisService("CNY and USD only fiat supported")

        url = "http://api.huobi.com/%smarket/detail_%s_json.js" % (
            fiat.lower(), crypto.lower()
        )
        r = self.get_url(url).json()
        return r['p_last']


class BTER(Service):
    service_id = 25
    api_homepage = "https://bter.com/api"
    name = "BTER"

    def fix_symbol(self, symbol):
        if symbol == 'bch':
            return 'bcc'
        return symbol

    def make_market(self, crypto, fiat):
        return ("%s_%s" % (self.fix_symbol(crypto), fiat)).lower()

    def get_current_price(self, crypto, fiat):
        url = "http://data.bter.com/api/1/ticker/%s" % self.make_market(crypto, fiat)
        response = self.get_url(url).json()
        if response.get('result', '') == 'false':
            raise ServiceError("BTER returned error: " + r['message'])
        return float(response['last'] or 0)

    def get_pairs(self):
        url = "http://data.bter.com/api/1/pairs"
        r = self.get_url(url).json()
        return [x.replace("_", "-") for x in r]

    def get_orderbook(self, crypto, fiat):
        url = "http://data.bter.com/api2/1/orderBook/%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'bids': [(float(x[0]), float(x[1])) for x in resp['bids']],
            'asks': [(float(x[0]), float(x[1])) for x in resp['asks']],
        }

    def _make_signature(self, params):
        return hmac.new(
            self.api_secret, urlencode(params), hashlib.sha512
        ).hexdigest()

    def _auth_request(self, url, params):
        raise Exception("Not tested")
        return self.post_url(url, headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Key': self.api_key,
            'Sign': self._make_signature(params)
        })

    def get_exchange_balance(self, currency, type="available"):
        url = "https://api.bter.com/api2/1/private/balances"
        resp = self._auth_request(url, {})
        for curr, bal in resp.json()[type].items():
            if curr == currency.upper():
                return float(bal)

    def get_deposit_address(self, currency):
        url = "https://bter.com/api2/1/private/depositAddress"
        resp = self._auth_request(url, {'currency': currency.upper()})
        return resp.json()['addr']


class Wex(Service):
    service_id = 7
    api_homepage = "https://wex.nz/api/documentation"
    name = "Wex"
    exchange_fee_rate = 0.002

    def check_error(self, response):
        try:
            j = response.json()
        except:
            raise ServiceError("Wex returned error: %s" % response.content)

        if 'error' in j:
            raise ServiceError("Wex returned error: %s" % j['error'])
        super(Wex, self).check_error(response)

    def make_market(self, crypto, fiat):
        return "%s_%s" % (
            self.fix_symbol(crypto).lower(),
            self.fix_symbol(fiat).lower()
        )

    def fix_symbol(self, symbol):
        if symbol == 'dash':
            return 'dsh'
        return symbol

    def reverse_fix_symbol(self, symbol):
        if symbol == 'dsh':
            return 'dash'
        return symbol

    def _fix_fiat_symbol(self, fiat):
        return fiat

    def get_current_price(self, crypto, fiat):
        pair = self.make_market(crypto, fiat)
        url = "https://wex.nz/api/3/ticker/" + pair
        response = self.get_url(url).json()
        return response[pair]['last']

    def get_pairs(self):
        url = "https://wex.nz/api/3/info"
        r = self.get_url(url).json()
        return [x.replace('_', '-') for x in r['pairs'].keys()]

    def get_orderbook(self, crypto, fiat):
        m = self.make_market(crypto, fiat)
        url = "https://wex.nz/api/3/depth/%s" % m
        resp = self.get_url(url).json()
        return {
            'bids': [(x[0], x[1]) for x in resp[m]['bids']],
            'asks': [(x[0], x[1]) for x in resp[m]['asks']]
        }

    def _make_signature(self, params):
        return hmac.new(
            self.api_secret, urlencode(params), hashlib.sha512
        ).hexdigest()

    def _auth_request(self, params):
        # max nonce wex will accept is 4294967294
        params['nonce'] = make_stateful_nonce(self.name)
        headers = {"Key": self.api_key, "Sign": self._make_signature(params)}
        return self.post_url("https://wex.nz/tapi", params, headers=headers)

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        params = {
            'method': 'Trade',
            'pair': self.make_market(crypto, fiat),
            'type': side,
            'rate': price,
            'amount': eight_decimal_places(amount),
        }
        return self._auth_request(params)
    make_order.supported_types = ['limit']
    make_order.minimums = {'btc': 0.001, 'ltc': 0.1}

    def get_deposit_address(self, crypto):
        params = {'coinName': crypto.lower(), 'method': 'CoinDepositAddress'}
        resp = self._auth_request(params).json()
        return resp['return']['address']

    def get_exchange_balance(self, currency, type="available"):
        resp = self._auth_request({'method': 'getInfo'}).json()
        try:
            return resp['return']['funds'][self.fix_symbol(currency).lower()]
        except IndexError:
            return 0

    def get_total_exchange_balances(self):
        resp = self._auth_request({'method': 'getInfo'}).json()['return']['funds']
        return {
            self.reverse_fix_symbol(code): bal for code, bal in resp.items()
            if not code.endswith("et") and bal > 0
        }

    def initiate_withdraw(self, currency, amount, address):
        resp = self._auth_request({
            'method': 'WithdrawCoin',
            'coinName': self.fix_symbol(currency),
            'amount': amount,
            'address': address,
        })
        return resp.json()


class ViaBTC(Service):
    service_id = 116

    def get_current_price(self, crypto, fiat):
        url = "https://www.viabtc.com/api/v1/market/ticker?market=%s%s" % (
            crypto.upper(), fiat.upper()
        )
        return float(self.get_url(url).json()['data']['ticker']['last'])


class CryptoDao(Service):
    service_id = 115
    api_homepage = "https://cryptodao.com/doc/api"

    def get_current_price(self, crypto, fiat):
        url = "https://cryptodao.com/api/ticker?source=%s&target=%s" % (
            fiat.upper(), crypto.upper()
        )
        r = self.get_url(url).json()
        return r['last']

    def get_orderbook(self, crypto, fiat):
        url = "https://cryptodao.com/api/depth?source=%s&target=%s" % (
            fiat.upper(), crypto.upper()
        )
        resp = self.get_url(url).json()
        return resp


class HitBTC(Service):
    service_id = 109
    api_homepage = "https://hitbtc.com/api"
    exchange_fee_rate = 0.001

    def check_error(self, response):
        j = response.json()

        if response.status_code == 400 and 'error' in j:
            e = j['error']
            raise SkipThisService("HitBTC returned %s %s: %s" % (
                e['code'], e['message'], e['description']
            ))

        if 'code' in j:
            raise SkipThisService("HitBTC returned %s: %s" % (j['code'], j['message']))
        super(HitBTC, self).check_error(response)

    def fix_symbol(self, symbol):
        return symbol.lower()

    def make_market(self, crypto, fiat):
        return ("%s%s" % (self.fix_symbol(crypto), self.fix_symbol(fiat))).upper()

    def get_pairs(self):
        url = 'https://api.hitbtc.com/api/1/public/symbols'
        r = self.get_url(url).json()['symbols']
        return [("%s-%s" % (x['commodity'], x['currency'])).lower() for x in r]

    def get_current_price(self, crypto, fiat):
        url = "https://api.hitbtc.com/api/1/public/%s/ticker" % self.make_market(crypto, fiat)
        r = self.get_url(url).json()
        return float(r['last'])

    def get_orderbook(self, crypto, fiat):
        url = "https://api.hitbtc.com/api/1/public/%s/orderbook" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'asks': [(float(x[0]), float(x[1])) for x in resp['asks']],
            'bids': [(float(x[0]), float(x[1])) for x in resp['bids']]
        }

    def _auth_request(self, path, params, method="post"):
        path = path + "?" + urlencode({
            'nonce': make_standard_nonce(),
            'apikey': self.api_key
        })
        headers = {"X-Signature": self._make_signature(path, params)}
        return self._external_request(
            method, "https://api.hitbtc.com" + path,
            params, headers=headers
        )

    def _make_signature(self, path, params):
        msg = path + urlencode(params)
        return hmac.new(self.api_secret, msg, hashlib.sha512).hexdigest()

    def get_exchange_balance(self, currency, type="available"):
        resp = self._auth_request("/api/1/trading/balance", {}, method="get").json()
        c = self.fix_symbol(currency).upper()
        try:
            matched = [x for x in resp['balance'] if x['currency_code'] == c][0]
        except IndexError:
            return 0

        if type == 'available':
            return float(matched['cash'])

        raise NotImplemented()

    def get_total_exchange_balances(self):
        resp = self._auth_request("/api/1/trading/balance", {}, method="get")
        return {
            self.fix_symbol(x['currency_code']): float(x['cash'])
            for x in resp.json()['balance'] if float(x['cash']) > 0
        }

    def get_deposit_address(self, currency):
        path = "/api/1/payment/address/%s" % self.fix_symbol(currency).upper()
        resp = self._auth_request(path, {}, method="get").json()
        return resp['address']

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        path = "/api/1/trading/new_order"
        import random, string
        clientOrderId = "".join(random.choice(string.digits + string.ascii_lowercase) for _ in range(30))
        params = {
            'symbol': self.make_market(crypto, fiat),
            'side': side,
            'price': price,
            'quantity': eight_decimal_places(amount),
            'clientOrderId': clientOrderId
        }
        if type == 'fill-or-kill':
            params['timeInForce'] = 'FOK'

        resp = self._auth_request(path, params).json()
        return resp
    make_order.minimums = {}
    make_order.supported_types = ['fill-or-kill', 'limit']


class Liqui(Service):
    service_id = 106

    def parse_market(self, market):
        crypto, fiat = market.split("_")
        if fiat == 'usdt':
            fiat = 'usd'
        return crypto, fiat

    def get_pairs(self):
        url = "https://api.liqui.io/api/3/info"
        r = self.get_url(url).json()['pairs']
        return ["%s-%s" % self.parse_market(x) for x in r.keys()]

    def make_market(self, crypto, fiat):
        return "%s_%s" % (
            self.fix_symbol(crypto).lower(), self.fix_symbol(fiat).lower()
        )

    def fix_symbol(self, symbol):
        if symbol == 'usd':
            return 'usdt'
        return symbol

    def get_current_price(self, crypto, fiat):
        pair = self.make_market(crypto, fiat)
        url = "https://api.liqui.io/api/3/ticker/%s" % pair
        return self.get_url(url).json()[pair]['last']

    def get_orderbook(self, crypto, fiat):
        pair = self.make_market(crypto, fiat)
        url = "https://api.liqui.io/api/3/depth/%s" % pair
        return self.get_url(url).json()[pair]

    def _make_signature(self, params):
        return hmac.new(
            self.api_secret, "?" + urlencode(params), hashlib.sha512
        ).hexdigest()

    def _auth_request(self, params):
        params['nonce'] = make_standard_nonce(small=True)
        headers = {
            'Key':self.api_key,
            'Sign': self._make_signature(params)
        }
        return self.post_url('https://api.liqui.io', params, headers=headers)

    def get_exchange_balance(self, currency):
        resp = self._auth_request({'method': 'getInfo'}).json()
        return resp

    def list_orders(self, crypto=None, fiat=None):
        resp = self._auth_request({'method': 'ActiveOrders'}).json()
        return resp

class CoinOne(Service):
    service_id = 105

    def get_pairs(self):
        return ['btc-krw', 'eth-krw', 'xrp-krw', 'etc-krw']

    def get_current_price(self, crypto, fiat):
        url = "https://api.coinone.co.kr/ticker?currency=%s" % crypto.lower()
        return float(self.get_url(url).json()['last'])


class CryptoBG(Service):
    service_id = 102

    def get_current_price(self, crypto, fiat):
        url = "https://crypto.bg/api/v1/public_rates"
        if crypto != 'btc' or fiat != 'bgn':
            raise SkipThisService("Only btc-bgn supported")
        return float(self.get_url(url).json()['rates']['bitcoin']['bid'])


class Bitso(Service):
    service_id = 101

    def get_current_price(self, crypto, fiat):
        url = "https://api.bitso.com/v3/ticker/?book=%s_%s" % (crypto, fiat)
        r = self.get_url(url.lower()).json()
        return float(['payload']['last'])

    def get_pairs(self):
        url = "https://api.bitso.com/v3/available_books/"
        r = self.get_url(url).json()['payload']
        return [x['book'].replace("_", '-') for x in r]


class TradeSatoshi(Service):
    service_id = 96

    def get_pairs(self):
        url = "https://tradesatoshi.com/api/public/getmarketsummaries"
        r = self.get_url(url).json()
        return [x['market'].replace("_", '-').lower() for x in r['result']]

    def get_current_price(self, crypto, fiat):
        url = "https://tradesatoshi.com/api/public/getticker?market=%s_%s" % (
            crypto.upper(), fiat.upper()
        )
        return self.get_url(url).json()['result']['last']


class UseCryptos(Service):
    service_id = 95

    def get_pairs(self):
        url = "https://usecryptos.com/jsonapi/pairs"
        r = self.get_url(url).json()
        return r

    def get_current_price(self, crypto, fiat):
        pair = "%s-%s" % (crypto.lower(), fiat.lower())
        url = "https://usecryptos.com/jsonapi/ticker/%s" % pair
        return self.get_url(url).json()['lastPrice']


class BitcoinIndonesia(Service):
    service_id = 94
    api_homepage = "https://blog.bitcoin.co.id/wp-content/uploads/2014/03/API-Documentation-Bitcoin.co_.id_.pdf"

    def get_current_price(self, crypto, fiat):
        url = "https://vip.bitcoin.co.id/api/%s_%s/ticker" % (crypto.lower(), fiat.lower())
        return float(self.get_url(url).json()['ticker']['last'])

class Kraken(Service):
    service_id = 93

    def check_error(self, response):
        if response.json()['error']:
            raise ServiceError("Kraken returned error: %s" % response.json()['error'][0])

        super(Kraken, self).check_error(response)

    def get_pairs(self):
        url = "https://api.kraken.com/0/public/AssetPairs"
        r = self.get_url(url).json()['result']
        ret = []
        for name, data in r.items():
            crypto = data['base'].lower()
            if len(crypto) == 4 and crypto.startswith('x'):
                crypto = crypto[1:]
            fiat = data['quote'].lower()
            if fiat.startswith("z"):
                fiat = fiat[1:]
            if crypto == 'xbt':
                crypto = 'btc'
            if fiat == 'xxbt':
                fiat = 'btc'
            if fiat == 'xeth':
                fiat = 'eth'
            ret.append("%s-%s" % (crypto, fiat))
        return list(set(ret))

    def get_current_price(self, crypto, fiat):
        if crypto != 'bch':
            crypto = "x" + crypto.lower()
            if crypto == 'xbtc':
                crypto = 'xxbt'
            fiat = "z" + fiat.lower()
            if fiat == 'zbtc':
                fiat = 'xxbt'
        else:
            # bch pairs have completely different format for some reason
            # my god kraken's api is terrible
            if fiat.lower() == 'btc':
                fiat = 'xbt'

        pair = "%s%s" % (crypto.upper(), fiat.upper())
        url = "https://api.kraken.com/0/public/Ticker?pair=%s" % pair
        r = self.get_url(url).json()['result']
        return float(r[pair]['c'][0])


class BTC38(Service):
    service_id = 92
    api_homepage = "http://www.btc38.com/help/document/2581.html"

    def get_current_price(self, crypto, fiat):
        url = 'http://api.btc38.com/v1/ticker.php?c=%s&mk_type=%s' % (crypto, fiat)
        return self.get_url(url).json()['ticker']['last']

    def get_pairs(self):
        url = "http://api.btc38.com/v1/ticker.php?c=all&mk_type=cny"
        cny_bases = self.get_url(url).json().keys()

        url = "http://api.btc38.com/v1/ticker.php?c=all&mk_type=btc"
        btc_bases = self.get_url(url).json().keys()

        return ["%s-cny" % x for x in cny_bases] + ["%s-btc" % x for x in btc_bases]


class BleuTrade(Service):
    service_id = 91
    api_homepage = 'https://bleutrade.com/help/API'

    def get_pairs(self):
        url = 'https://bleutrade.com/api/v2/public/getmarkets'
        r = self.get_url(url).json()['result']
        return [x['MarketName'].lower().replace("_", '-') for x in r]

    def get_current_price(self, crypto, fiat):
        url = "https://bleutrade.com/api/v2/public/getticker?market=%s_%s" % (
            crypto.upper(), fiat.upper()
        )
        r = self.get_url(url).json()
        return float(r['result'][0]['Last'])



class xBTCe(Service):
    service_id = 90
    name = "xBTCe"
    api_homepage = "https://www.xbtce.com/tradeapi"

    def get_current_price(self, crypto, fiat):
        if crypto.lower() == 'dash':
            crypto = "dsh"
        if fiat.lower() == 'rur':
            fiat = 'rub'
        if fiat.lower() == 'cny':
            fiat = 'cnh'
        pair = "%s%s" % (crypto.upper(), fiat.upper())
        url = "https://cryptottlivewebapi.xbtce.net:8443/api/v1/public/ticker/%s" % pair
        r = self.get_url(url).json()
        try:
            return r[0]['LastSellPrice']
        except IndexError:
            raise ServiceError("Pair not found")

    def get_pairs(self):
        url = "https://cryptottlivewebapi.xbtce.net:8443/api/v1/public/symbol"
        r = self.get_url(url).json()
        ret = []
        for pair in r:
            crypto = pair['MarginCurrency'].lower()
            fiat = pair['ProfitCurrency'].lower()

            if crypto.lower() == 'dsh':
                crypto = "dash"
            if fiat.lower() == 'rub':
                fiat = 'rur'
            if fiat == 'cnh':
                fiat = 'cny'
            ret.append(("%s-%s" % (crypto, fiat)))

        return list(set(ret))


class Cryptopia(Service):
    service_id = 82
    api_homepage = "https://www.cryptopia.co.nz/Forum/Thread/255"
    exchange_fee_rate = 0.002

    def check_error(self, response):
        r = response.json()
        error = r.get('Error', False)
        if error is not None:
            raise ServiceError("Cryptopia returned error: %s" % r['Error'])
        super(Cryptopia, self).check_error(response)

    def fix_symbol(self, symbol):
        if symbol.lower() in ['nzd', 'usd']:
            symbol += "t"
        return symbol

    def reverse_fix_symbol(self, symbol):
        symbol = symbol.lower()
        if symbol in ['nzdt', 'usdt']:
            symbol = symbol[:-1]
        return symbol

    def make_market(self, crypto, fiat):
        return "%s_%s" % (
            self.fix_symbol(crypto).upper(), self.fix_symbol(fiat).upper()
        )

    def get_current_price(self, crypto, fiat):
        url = "https://www.cryptopia.co.nz/api/GetMarket/%s" % self.make_market(crypto, fiat)
        r = self.get_url(url).json()
        return r['Data']['LastPrice']

    def get_pairs(self):
        url = "https://www.cryptopia.co.nz/api/GetTradePairs"
        r = self.get_url(url).json()['Data']
        ret = []
        for pair in r:
            crypto = pair['Symbol']
            fiat = pair['BaseSymbol']
            if fiat.lower() == 'usdt':
                fiat = 'usd'
            ret.append(("%s-%s" % (crypto, fiat)).lower())

        return ret

    def get_orderbook(self, crypto, fiat):
        url = "https://www.cryptopia.co.nz/api/GetMarketOrders/%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'bids': [(x['Price'], x['Total']) for x in resp['Data']['Buy']],
            'asks': [(x['Price'], x['Total']) for x in resp['Data']['Sell']]
        }


    def _make_signature_header(self, url, params):
        nonce = str(int(time.time()))
        post_data = json.dumps(params);
        m = hashlib.md5()
        m.update(post_data)
        requestContentBase64String = base64.b64encode(m.digest())
        signature = self.api_key + "POST" + quote_plus(url).lower() + nonce + requestContentBase64String
        hmacsignature = base64.b64encode(
            hmac.new(base64.b64decode(self.api_secret), signature, hashlib.sha256).digest()
        )
        header_value = "amx " + self.api_key + ":" + hmacsignature + ":" + nonce
        return {'Authorization': header_value, 'Content-Type':'application/json; charset=utf-8' }

    def _auth_request(self, method, args):
        url = "https://www.cryptopia.co.nz/Api/" + method
        return self.post_url(url, json=args, headers=self._make_signature_header(url, args))

    def make_order(self, crypto, fiat, amount, price, type="limit", side="buy"):
        args = {
            'Market': ("%s/%s" % (self.fix_symbol(crypto), self.fix_symbol(fiat))).upper(),
            'Type': side,
            'Rate': price,
            'Amount': eight_decimal_places(amount)
        }
        resp = self._auth_request("SubmitTrade", args).json()
        return resp['Data']['OrderId']
    make_order.minimums = {}
    make_order.supported_types = ['limit']

    def get_exchange_balance(self, currency):
        curr = self.fix_symbol(currency).upper()
        try:
            resp = self._auth_request('GetBalance', {'Currency': curr})
        except ServiceError:
            return 0
        for item in resp.json()['Data']:
            if item['Symbol'] == curr:
                return item['Total']

    def get_total_exchange_balances(self):
        resp = self._auth_request('GetBalance', {}).json()
        #return resp.json()
        return {
            self.reverse_fix_symbol(x['Symbol']): x['Total'] for x in resp['Data']
            if x['Total'] > 0
        }

    def get_deposit_address(self, currency):
        curr = self.fix_symbol(currency).upper()
        resp = self._auth_request('GetDepositAddress', {'Currency': curr})
        return resp.json()['Data']['Address']

    def initiate_withdraw(self, currency, amount, address):
        curr = self.fix_symbol(currency).upper()
        resp = self._auth_request('SubmitWithdraw', {
            'Currency': curr,
            'Address': address,
            'Amount': amount
        })
        return resp.json()

class YoBit(Service):
    service_id = 77
    api_homepage = "https://www.yobit.net/en/api/"

    def get_current_price(self, crypto, fiat):
        pair = "%s_%s" % (crypto.lower(), fiat.lower())
        url = "https://yobit.net/api/3/ticker/%s" % pair
        r = self.get_url(url).json()

        if 'error' in r:
            raise SkipThisService(r['error'])

        return r[pair]['last']

    def get_pairs(self):
        url = 'https://yobit.net/api/3/info'
        r = self.get_url(url).json()
        return [x.replace("_", '-') for x in r['pairs'].keys()]


class Yunbi(Service):
    service_id = 78
    api_homepage = "https://yunbi.com/swagger"

    def get_current_price(self, crypto, fiat):
        if fiat.lower() != "cny":
            raise SkipThisService("Only CNY markets supported")
        url = "https://yunbi.com/api/v2/tickers/%s%s.json" % (crypto.lower(), fiat.lower())
        r = self.get_url(url, headers={"Accept": "application/json"}).json()
        return float(r['ticker']['last'])

    def get_pairs(self):
        url = "https://yunbi.com/api/v2/markets.json"
        r = self.get_url(url).json()
        ret = []
        for pair in r:
            ret.append(pair['name'].replace("/", '-').lower())
        return ret


class Vircurex(Service):
    service_id = 70
    base_url = "https://api.vircurex.com/api"
    api_homepage = "https://vircurex.com/welcome/api"

    def check_error(self, response):
        j = response.json()
        if j['status'] != 0:
            raise ServiceError("Vircurex returned error: %s" % j['status_text'])

        super(Vircurex, self).check_error(response)

    def get_current_price(self, crypto, fiat):
        if crypto == 'blk':
            crypto = 'bc'
        url = "%s/get_last_trade.json?base=%s&alt=%s" % (
            self.base_url, crypto.upper(), fiat.upper()
        )
        r = self.get_url(url).json()
        return float(r['value'])

    def get_pairs(self):
        url = "%s/get_info_for_currency.json" % self.base_url
        r = self.get_url(url).json()
        ret = []
        for fiat, data in r.items():
            if fiat == 'status':
                continue
            for crypto, exchange_data in data.items():
                pair = "%s-%s" % (crypto.lower(), fiat.lower())
                ret.append(pair)
        return ret


class LiveCoin(Service):
    service_id = 110
    base_url = "https://api.livecoin.net"
    api_homepage = "https://www.livecoin.net/api/public"

    def get_pairs(self):
        url = "%s/exchange/ticker" % (self.base_url)
        r = self.get_url(url).json()
        return [x['symbol'].replace('/', '-').lower() for x in r]

    def get_current_price(self, crypto, fiat):
        url = "%s/exchange/ticker/?currencyPair=%s/%s" % (
            self.base_url, crypto.upper(), fiat.upper()
        )
        return self.get_url(url).json()['last']

class Bithumb(Service):
    service_id = 129

    def get_current_price(self, crypto, fiat):
        if fiat != 'krw':
            raise SkipThisService("Only KRW supported")
        url = "https://api.bithumb.com/public/ticker/%s" % crypto.upper()
        resp = self.get_url(url).json()
        return float(resp['data']['average_price'])

    def get_orderbook(self, crypto, fiat):
        if fiat != 'krw':
            raise SkipThisService("Only KRW supported")

        url = "https://api.bithumb.com/public/orderbook/%s" % crypto
        resp =  self.get_url(url).json()['data']
        return {
            'bids': [(float(x['price']), float(x['quantity'])) for x in resp['bids']],
            'asks': [(float(x['price']), float(x['quantity'])) for x in resp['asks']]
        }

class Binance(Service):
    service_id = 130

    def check_error(self, response):
        j = response.json()
        if 'code' in j:
            raise ServiceError("Binance returned error: %s %s" % (j['code'], j['msg']))
        super(Binance, self).check_error(response)

    def make_market(self, crypto, fiat):
        return ("%s%s" % (self.fix_symbol(crypto), self.fix_symbol(fiat))).upper()

    def parse_market(self, market):
        market = market.lower()
        if market.endswith("usdt"):
            crypto, fiat = market[:-4], "usd"
        elif market.endswith("eth"):
            crypto, fiat = market[:-3], "eth"
        elif market.endswith("btc"):
            crypto, fiat = market[:-3], "btc"
        else:
            crypto, fiat = market[:-3], market[-3:]

        if crypto == 'iota':
            crypto = 'miota'

        if crypto == 'bcc':
            crypto = 'bch'

        return "%s-%s" % (crypto, fiat)

    def fix_symbol(self, symbol):
        if symbol.lower() == 'usd':
            return 'usdt'
        if symbol == 'miota':
            return 'iota'
        if symbol == 'bch':
            return "bcc"
        return symbol

    def get_current_price(self, crypto, fiat):
        url = "https://www.binance.com/api/v1/ticker/allPrices"
        resp = self.get_url(url).json()
        for data in resp:
            if data['symbol'] == self.make_market(crypto, fiat):
                return float(data['price'])
        raise SkipThisService("Market not found")

    def get_pairs(self):
        url = "https://www.binance.com/api/v1/ticker/allPrices"
        resp = self.get_url(url).json()
        symbols = []
        for data in resp:
            symbols.append(self.parse_market(data['symbol']))
        return symbols

    def get_orderbook(self, crypto, fiat):
        url = "https://www.binance.com/api/v1/depth"
        resp = self.get_url(url, {'symbol': self.make_market(crypto, fiat)}).json()
        return {
            'bids': [(float(x[1]), float(x[0])) for x in resp['bids']],
            'asks': [(float(x[1]), float(x[0])) for x in resp['asks']]
        }

    def _auth_request(self, path, params, method="post"):
        params['timestamp'] = make_standard_nonce()
        params['signature'] = self._make_signature(params)
        headers = {"X-MBX-APIKEY": self.api_key}
        return self._external_request(
            method, "https://www.binance.com" + path, params, headers=headers
        )

    def _make_signature(self, params):
        return hmac.new(
            self.api_secret, urlencode(params), hashlib.sha256
        ).hexdigest()

    def get_exchange_balance(self, currency, type="available"):
        if type == 'available':
            type = 'free'
        else:
            type == 'locked'
        path = "/api/v3/account"
        resp = self._auth_request(path, {}, method="get").json()
        for data in resp['balances']:
            if data['asset'].lower() == currency.lower():
                return float(data[type])

        return 0


class BitFlyer(Service):
    service_id = 111
    api_homepage = "https://bitflyer.jp/API?top_link&footer"

    def get_current_price(self, crypto, fiat):
        url = "https://api.bitflyer.jp/v1/getticker?product_code=%s" % (
            self.make_market(crypto, fiat)
        )
        r = self.get_url(url).json()
        return r['ltp']

    def get_pairs(self):
        return ['btc-jpy', 'eth-btc', 'btc-usd']

    def make_market(self, crypto, fiat):
        return ("%s_%s" % (crypto, fiat)).upper()

    def get_orderbook(self, crypto, fiat):
        if fiat.lower() == 'jpy':
            domain = "api.bitflyer.jp"
        elif fiat.lower() == 'usd':
            domain = "api.bitflyer.com"
        else:
            raise SkipThisService("Only jpy and usd suppported")
        url = "https://%s/v1/getboard?product_code=%s" % (
            domain, self.make_market(crypto, fiat)
        )
        resp = self.get_url(url).json()
        return {
            'bids': [(x['price'], x['size']) for x in resp['bids']],
            'asks': [(x['price'], x['size']) for x in resp['asks']],
        }

    def get_block(self, crypto, block_number=None, block_hash=None, latest=False):
        url = "https://chainflyer.bitflyer.jp/v1/block/%s" % (
            block_hash or
            ('height/%s' % block_number if block_number else None) or
            ('latest' if latest else 'None')
        )
        r = self.get_url(url).json()
        return dict(
            block_number=r['height'],
            time=arrow.get(r['timestamp']).datetime,
            #mining_difficulty=r['difficulty'],
            hash=r['block_hash'],
            next_hash=r.get('nextblockhash', None),
            previous_hash=r.get('prev_block'),
            txids=r['tx_hashes'],
            version=r['version']
        )

class BitX(Service):
    service_id = 131
    api_homepage = "https://www.luno.com/en/api"

    def parse_market(self, market):
        if market.startswith("XBT"):
            crypto = "BTC"
            fiat = market[3:]
        return crypto, fiat

    def fix_symbol(self, symbol):
        if symbol.lower() == 'btc':
            return 'XBT'
        return symbol

    def make_market(self, crypto, fiat):
        return ("%s%s" % (self.fix_symbol(crypto), self.fix_symbol(fiat))).upper()

    def get_current_price(self, crypto, fiat):
        url = "https://api.mybitx.com/api/1/ticker?pair=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return float(resp['last_trade'])

    def get_pairs(self):
        url = "https://api.mybitx.com/api/1/tickers"
        resp = self.get_url(url).json()['tickers']
        return [
            ("%s-%s" % self.parse_market(x['pair'])).lower() for x in resp
        ]

    def get_orderbook(self, crypto, fiat):
        url = "https://api.mybitx.com/api/1/orderbook?pair=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'bids': [(float(x['price']), float(x['volume'])) for x in resp['bids']],
            'asks': [(float(x['price']), float(x['volume'])) for x in resp['asks']]
        }

class ItBit(Service):
    service_id = 132

    def fix_symbol(self, symbol):
        if symbol.lower() == 'btc':
            return 'XBT'
        return symbol

    def make_market(self, crypto, fiat):
        return ("%s%s" % (self.fix_symbol(crypto), self.fix_symbol(fiat))).upper()

    def get_current_price(self, crypto, fiat):
        url = "https://api.itbit.com/v1/markets/%s/ticker" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return float(resp['lastPrice'])

    def get_pairs(self):
        return ['btc-usd', 'btc-sgd', 'btc-eur']

    def get_orderbook(self, crypto, fiat):
        url = "https://api.itbit.com/v1/markets/%s/order_book" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'bids': [(float(x[0]), float(x[1])) for x in resp['bids']],
            'asks': [(float(x[0]), float(x[1])) for x in resp['asks']]
        }

    def _make_signature(self):
        pass # https://github.com/itbit/itbit-restapi-python/blob/master/itbit_api.py

class KuCoin(Service):
    service_id = 133
    symbol_mapping = (
        ('usd', 'usdt'),
    )

    def make_market(self, crypto, fiat):
        return ("%s-%s" % (self.fix_symbol(crypto), self.fix_symbol(fiat))).upper()

    def parse_market(self, market):
        return super(KuCoin, self).parse_market(market.lower(), '-')

    def get_pairs(self):
        url = "https://api.kucoin.com/v1/market/open/symbols"
        resp = self.get_url(url).json()
        pairs = []
        for pair in resp['data']:
            crypto, fiat = self.parse_market(pair['symbol'])
            pairs.append("%s-%s" % (crypto, fiat))
        return pairs

    def get_orderbook(self, crypto, fiat):
        url = "https://api.kucoin.com/v1/open/orders?symbol=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()['data']
        return {
            'bids': [(x[0], x[1]) for x in resp['BUY']],
            'asks': [(x[0], x[1]) for x in resp['SELL']]
        }

    def get_current_price(self, crypto, fiat):
        url = "https://api.kucoin.com/v1/open/tick?symbol=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()['data']
        return resp['lastDealPrice']

class CCex(Service):
    service_id = 134
    api_homepage = "https://c-cex.com/?id=api"
    base_url = "https://c-cex.com/t/api_pub.html"

    def check_error(self, response):
        if response.content == "Maintenance...":
            raise ServiceError("C-Cex is down for maintenance")
        super(CCex, self).check_error(response)

    def make_market(self, crypto, fiat):
        return "%s-%s" % (crypto.lower(), fiat.lower())

    def get_current_price(self, crypto, fiat):
        url = "https://c-cex.com/t/%s.json" % (
            self.make_market(crypto, fiat)
        )
        response = self.get_url(url).json()
        return float(response['ticker']['lastprice'])

    def get_pairs(self):
        url = "https://c-cex.com/t/pairs.json"
        r = self.get_url(url).json()
        return r['pairs']

    def get_orderbook(self, crypto, fiat):
        url = "%s?a=getorderbook&market=%s&type=both" % (
            self.base_url, self.make_market(crypto, fiat)
        )
        resp = self.get_url(url).json()['result']
        return {
            'bids': [(x["Rate"], x["Quantity"]) for x in resp['buy']],
            'asks': [(x["Rate"], x["Quantity"]) for x in resp['sell']]
        }

    def _auth_request(self, params):
        params['nonce'] = make_standard_nonce()
        params['apikey'] = self.api_key
        url = "%s?%s" % (self.base_url, urlencode(params))
        headers = {'apisign': self._make_signature(url)}
        return self.get_url(url, headers=headers)

    def _make_signature(self, url):
        return hmac.new(
            self.api_secret,
            msg=url,
            digestmod=hashlib.sha512
        ).hexdigest()

    def get_exchange_balance(self, currency, type="available"):
        resp = self._auth_request({'a': 'getbalance', 'currency': currency})
        if type == 'available':
            return resp.json()['Available']

    def get_deposit_address(self, currency):
        resp = self._auth_request({'a': 'getbalance', 'currency': currency})
        return resp.json()['CryptoAddress']

class CoinEx(Service):
    service_id = 138

    def __init__(self, access_id=None, **kwargs):
        self.access_id = access_id
        return super(CoinEx, self).__init__(**kwargs)

    def get_pairs(self):
        url = "https://api.coinex.com/v1/market/list"
        resp = self.get_url(url).json()['data']
        return [("%s-%s" % (x[:-3], x[-3:])).lower() for x in resp]

    def make_market(self, crypto, fiat):
        return ("%s%s" % (crypto, fiat)).upper()

    def get_current_price(self, crypto, fiat):
        url = "https://api.coinex.com/v1/market/ticker?market=%s" % (
            self.make_market(crypto, fiat)
        )
        resp = self.get_url(url).json()
        return float(resp['data']['ticker']['last'])

    def _auth_request(self, url, params=None):
        params['tonce'] = make_standard_nonce()
        params['access_id'] = self.access_id
        str_params = urlencode(sorted(params.items(), key=lambda x: x[0]))
        to_sign = str_params + "&secret_key=%s" % self.api_secret
        digest = hashlib.md5(to_sign).hexdigest().upper()
        return self.get_url(url + "?" + str_params, headers={
            'Content-Type': 'application/json',
            'authorization': digest
        })

    def get_exchange_balance(self, crypto):
        url = "https://api.coinex.com/v1/balance/"
        resp = self._auth_request(url, {}).json()
        return resp


class OKEX(Service):
    service_id = 139
    api_homepage = 'https://www.okex.com/rest_api.html'
    symbol_mapping = (
        ('usd', 'usdt'),
    )

    def check_error(self, response):
        j = response.json()
        if 'error_code' in j:
            raise ServiceError("OKEX returned error: %s" % j['error_code'])
        super(OKEX, self).check_error(response)

    def get_current_price(self, crypto, fiat):
        url = "https://www.okex.com/api/v1/ticker.do?symbol=%s" % (
            self.make_market(crypto, fiat)
        )
        resp = self.get_url(url).json()
        return float(resp['ticker']['last'])

class BitZ(Service):
    service_id = 140

    def check_error(self, response):
        j = response.json()
        if not j['code'] == 0:
            raise ServiceError("BitZ returned error: %s: %s" % (
                j['code'], j['msg']
            ))
        super(BitZ, self).check_error(response)

    def get_current_price(self, crypto, fiat):
        url = "https://www.bit-z.com/api_v1/ticker?coin=%s" % (self.make_market(crypto, fiat))
        resp = self.get_url(url).json()
        return float(resp['data']['last'])

class Zaif(Service):
    service_id = 141

    def check_error(self, response):
        j = response.json()
        if 'error' in j:
            raise ServiceError("Zaif returned error: %s" % (j['error']))
        super(Zaif, self).check_error(response)

    def get_current_price(self, crypto, fiat):
        url = "https://api.zaif.jp/api/1/ticker/%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return resp['last']

class Korbit(Service):
    service_id = 142
    api_homepage = "https://apidocs.korbit.co.kr/"

    def get_current_price(self, crypto, fiat):
        url = "https://api.korbit.co.kr/v1/ticker?currency_pair=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return float(resp['last'])

    def get_orderbook(self, crypto, fiat):
        url = "https://api.korbit.co.kr/v1/orderbook?currency_pair=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'asks': [(float(x[0]), float(x[1])) for x in resp['asks']],
            'bids': [(float(x[0]), float(x[1])) for x in resp['bids']]
        }

class CoinEgg(Service):
    service_id = 143
    api_homepage = "https://www.coinegg.com/explain.api.html#partone"

    def get_current_price(self, crypto, fiat):
        if fiat.lower() != 'btc':
            raise SkipThisService("Only BTC markets supported")
        url = "https://api.coinegg.com/api/v1/ticker/?coin=%s" % crypto
        resp = self.get_url(url).json()
        return float(resp['last'])

    def get_orderbook(self, crypto, fiat):
        if fiat.lower() != 'btc':
            raise SkipThisService("Only BTC markets supported")

        url = "https://api.coinegg.com/api/v1/depth/"
        resp = self.get_url(url).json()
        return {
            'bids': [(float(x[0]), float(x[1])) for x in resp['bids']],
            'asks': [(float(x[0]), float(x[1])) for x in resp['asks']]
        }

class ZB(Service):
    service_id = 144
    api_homepage = "https://www.zb.com/i/developer"
    symbol_mapping = (
        ('usd', 'usdt'),
        ('bch', 'bcc')
    )

    def get_pairs(self):
        url = "http://api.zb.com/data/v1/markets"
        resp = self.get_url(url).json()
        pairs = []
        for pair in resp.keys():
            pairs.append("%s-%s" % self.parse_market(pair))
        return pairs

    def get_current_price(self, crypto, fiat):
        url = "http://api.zb.com/data/v1/ticker?market=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return float(resp['ticker']['last'])

    def get_orderbook(self, crypto, fiat):
        url = "http://api.zb.com/data/v1/depth?market=%s&size=3" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        del resp['timestamp']
        return resp

class CoinNest(Service):
    service_id = 145
    api_homepage = "https://www.coinnest.co.kr/doc/intro.html"

    def get_current_price(self, crypto, fiat):
        if fiat.lower() != 'krw':
            raise SkipThisService("Only KRW markets supported")
        url = "https://api.coinnest.co.kr/api/pub/ticker?coin=%s" % crypto
        resp = self.get_url(url).json()
        return resp['last']

    def get_orderbook(self, crypto, fiat):
        if fiat.lower() != 'krw':
            raise SkipThisService("Only KRW markets supported")
        url = "https://api.coinnest.co.kr/api/pub/depth?coin=%s" % crypto
        resp = self.get_url(url).json()
        del resp['result']
        return resp

class BitBank(Service):
    service_id = 147
    api_homepage = "https://docs.bitbank.cc/"
    symbol_mapping = (
        ('bch', 'bcc'),
    )

    def get_current_price(self, crypto, fiat):
        url = "https://public.bitbank.cc/%s/ticker" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return float(resp['data']['last'])

    def get_orderbook(self, crypto, fiat):
        url = "https://public.bitbank.cc/%s/depth" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'asks': [(float(x[0]), float(x[1])) for x in resp['data']['asks']],
            'bids': [(float(x[0]), float(x[1])) for x in resp['data']['bids']]
        }

class EXX(Service):
    service_id = 148
    symbol_mapping = (
        ('usd', 'usdt'),
        ('bch', 'bcc')
    )

    def get_pairs(self):
        url = "https://api.exx.com/data/v1/markets"
        resp = self.get_url(url).json()
        pairs = []
        for pair in resp.keys():
            pairs.append("%s-%s" % self.parse_market(pair))
        return pairs

    def get_current_price(self, crypto, fiat):
        url = "https://api.exx.com/data/v1/ticker?currency=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return float(resp['ticker']['last'])

    def get_orderbook(self, crypto, fiat):
        url = "https://api.exx.com/data/v1/depth?currency=%s" % self.make_market(crypto, fiat)
        resp = self.get_url(url).json()
        return {
            'asks': [(float(x[0]), float(x[1])) for x in resp['asks']],
            'bids': [(float(x[0]), float(x[1])) for x in resp['bids']]
        }



import datetime
from tabulate import tabulate
import pytz

from moneywagon.crypto_data import crypto_data
from moneywagon.core import CurrencyNotSupported, make_standard_halfing_eras
from moneywagon.blocktime_adjustments import adjustments

class SupplyEstimator(object):
    """
    Returns a function that can be used to calculate the supply of coins for a given
    amount of minutes since genesis date. Only works for standard bitcoin forked
    currencies such as LTC, DOGE, PPC, etc.
    """

    def __init__(self, crypto=None, supply_data=None, genesis_date=None, blocktime_adjustments=True):
        """
        Can be initilaized by either passing in a crypto symbol, or by passing in
        supply_data and genesis_date together. If blocktime_adjustments is True,
        the blocktime adjustments will be taken from the samples bundled with moneywagon.
        You can also pass in your own samples (using `get_block_adjustments`)
        """
        if crypto:
            try:
                cd = crypto_data[crypto.lower()]
            except KeyError:
                raise CurrencyNotSupported("Code '%s' not supported" % crypto)
            try:
                self.genesis_date = cd['genesis_date']
                self.supply_data = cd['supply_data']
            except KeyError:
                raise NotImplementedError("Insufficient supply data for %s" % crypto.upper())
        else:
            self.supply_data = supply_data
            self.genesis_date = genesis_date

        self.minutes_per_block = self.supply_data['minutes_per_block']
        self.method = self.supply_data['method']
        self.blocktime_adjustments = blocktime_adjustments
        if blocktime_adjustments is True:
            self.blocktime_adjustments = adjustments.get(crypto)

    def make_supply_table(self, supply_divide=1, table_format='simple'):
        eras = self.supply_data.get('eras')
        if not eras and self.supply_data['method'] == 'standard':
            eras = make_standard_halfing_eras(
                start=0, interval=self.supply_data['blocks_per_era'],
                start_reward=self.supply_data['start_coins_per_block']
            )

        tag = ""
        if supply_divide == 1e6:
            tag = " (in millions)"
        if supply_divide == 1e9:
            tag = " (in billions)"

        total_supply = (self.supply_data.get('full_cap') or 0) / supply_divide
        rows = []
        running_total = 0
        for era, data in enumerate(eras, 1):
            start = data['start']
            end = data['end']
            reward = data['reward']
            total = ((end - start) * reward) / supply_divide if end else ""
            running_total += total or 0
            percent = "%.2f" % (float(running_total * 100) / total_supply) if total_supply else None
            date = self.estimate_date_from_height(end) if end else None
            row = [era, start, "{0:%m-%d-%Y}".format(date) if date else "", end, reward, total, running_total]

            if total_supply:
                row.append(percent)

            rows.append(row)

        headers = [
            'Era', 'Start Block', 'End Date', 'End Block', 'Reward per block',
            'Total Created This Era' + tag, "Total Existing" + tag
        ]
        if total_supply:
            headers.append("Percentage Issued")

        return tabulate(rows, headers=headers, tablefmt=table_format)

    @property
    def block_adjustment_in_minutes(self):
        minute_adjustments = []
        minutes_since_genesis = 0
        previous_minutes_per_block = self.minutes_per_block
        previous_adjustment_block = 0
        for adjustment_block, new_minutes_per_block in self.blocktime_adjustments:
            minutes_since_last_adjustment = (adjustment_block - previous_adjustment_block) * previous_minutes_per_block
            #minutes_since_genesis += minutes_since_last_adjustment
            minute_adjustments.append([minutes_since_last_adjustment, new_minutes_per_block])

            previous_minutes_per_block = new_minutes_per_block
            previous_adjustment_block = adjustment_block

        return minute_adjustments

    def estimate_height_from_date(self, at_time):
        if not hasattr(at_time, 'hour'):
            at_time = datetime.datetime.fromordinal(at_time.toordinal())
        minutes = (at_time - self.genesis_date).total_seconds() / 60.0
        target_block = int(minutes / self.minutes_per_block)

        if self.blocktime_adjustments:
            new_block = 0
            previous_minutes_per_block = self.minutes_per_block
            minutes_since_genesis = 0

            for minutes_since_last_adjustment, new_minutes_per_block in self.block_adjustment_in_minutes:
                if minutes_since_last_adjustment < minutes:
                    this_blocks = minutes_since_last_adjustment / previous_minutes_per_block
                    new_block += this_blocks
                else:
                    remainder_blocks = (minutes - minutes_since_genesis) / previous_minutes_per_block
                    new_block += remainder_blocks
                    break

                previous_minutes_per_block = new_minutes_per_block
                minutes_since_genesis += minutes_since_last_adjustment
            else:
                remainder_blocks = (minutes - minutes_since_genesis) / new_minutes_per_block
                new_block += remainder_blocks

            target_block = new_block

        return target_block

    def estimate_date_from_height(self, block_height):
        minutes = block_height * self.minutes_per_block
        if self.blocktime_adjustments:
            new_minutes = 0
            previous_adjustment_block = 0
            previous_minutes_per_block = self.minutes_per_block
            for adjustment_block, new_minutes_per_block in self.blocktime_adjustments:
                if block_height > adjustment_block:
                    minutes_this_adjustment = (adjustment_block - previous_adjustment_block) * previous_minutes_per_block
                    new_minutes += minutes_this_adjustment
                else:
                    remaining_minutes = (block_height - previous_adjustment_block) * previous_minutes_per_block
                    new_minutes += remaining_minutes
                    break

                previous_minutes_per_block = new_minutes_per_block
                previous_adjustment_block = adjustment_block
            else:
                remaining_minutes = (block_height - previous_adjustment_block) * previous_minutes_per_block
                new_minutes += remaining_minutes

            minutes = new_minutes

        return self.genesis_date + datetime.timedelta(minutes=minutes)

    def estimate_confirmations(self, confirmed_at_time):
        confirmed_block = self.estimate_height_from_date(confirmed_at_time)
        current_block = self.estimate_height_from_date(datetime.datetime.now())
        return current_block - confirmed_block

    def calculate_supply(self, block_height=None, at_time=None):
        if at_time:
            block_height = self.estimate_height_from_date(at_time)

        if self.method == 'standard':
            return self._standard_supply(block_height)
        if self.method == 'per_era':
            return self._per_era_supply(block_height)

    def _per_era_supply(self, block_height):
        """
        Calculate the coin supply based on 'eras' defined in crypto_data. Some
        currencies don't have a simple algorithmically defined halfing schedule
        so coins supply has to be defined explicitly per era.
        """
        coins = 0
        for era in self.supply_data['eras']:
            end_block = era['end']
            start_block = era['start']
            reward = era['reward']

            if not end_block or block_height <= end_block:
                blocks_this_era = block_height - start_block
                coins += blocks_this_era * reward
                break

            blocks_per_era = end_block - start_block
            coins += reward * blocks_per_era

        return coins

    def _standard_supply(self, block_height):
        """
        Calculate the supply of coins for a given time (in either datetime, or
        block height) for coins that use the "standard" method of halfing.
        """
        start_coins_per_block = self.supply_data['start_coins_per_block']
        minutes_per_block = self.supply_data['minutes_per_block']
        blocks_per_era = self.supply_data['blocks_per_era']
        full_cap = self.supply_data.get('full_cap')

        if not full_cap:
            full_cap = 100000000000 # nearly infinite

        coins = 0
        for era, start_block in enumerate(range(0, full_cap, blocks_per_era), 1):
            end_block = start_block + blocks_per_era
            reward = start_coins_per_block / float(2 ** (era - 1))
            if block_height < end_block:
                blocks_this_era = block_height - start_block
                coins += blocks_this_era * reward
                break

            coins += reward * blocks_per_era

        return coins

def get_block_adjustments(crypto, points=None, intervals=None, **modes):
    """
    This utility is used to determine the actual block rate. The output can be
    directly copied to the `blocktime_adjustments` setting.
    """
    from moneywagon import get_block
    all_points = []

    if intervals:
        latest_block_height = get_block(crypto, latest=True, **modes)['block_number']
        interval = int(latest_block_height / float(intervals))
        all_points = [x * interval for x in range(1, intervals - 1)]

    if points:
        all_points.extend(points)

    all_points.sort()

    adjustments = []
    previous_point = 0
    previous_time = (crypto_data[crypto.lower()].get('genesis_date').replace(tzinfo=pytz.UTC)
        or get_block(crypto, block_number=0, **modes)['time']
    )

    for point in all_points:
        if point == 0:
            continue
        point_time = get_block(crypto, block_number=point, **modes)['time']
        length = point - previous_point
        minutes = (point_time - previous_time).total_seconds() / 60
        rate = minutes / length
        adjustments.append([previous_point, rate])

        previous_time = point_time
        previous_point = point

    return adjustments

def get_block_currencies():
    """
    Returns a list of all curencies (by code) that have a service defined that
    implements `get_block`.
    """
    return ['btc', 'ltc', 'ppc', 'dash', 'doge', 'ric']
    currencies = []
    for currency, data in crypto_data.items():
        if type(data) is list:
            continue
        block_services = data.get('services', {}).get('get_block', [])
        if len(block_services) > 0 and not all([x.get_block.by_latest for x in block_services]):

            currencies.append(currency)

    return currencies

def write_blocktime_adjustments(path, **modes):
    with open(path, "w") as f:
        f.write("# last updated: %s\n\nadjustments = {\n" % datetime.datetime.now())
        for currency in get_block_currencies():
            if modes.get('verbose'): print("getting adjustments for %s" % currency)
            try:
                points = crypto_data[currency]['supply_data'].get('additional_block_interval_adjustment_points', [])
                adjustments = get_block_adjustments(currency, points=points, intervals=25, **modes)
                f.write("'%s': [\n%s],\n" % (currency, ''.join(['    %s,\n' % x for x in adjustments])))
            except Exception as exc:
                print("broken", currency, exc)

        f.write("}")
